import logging
import os
import sys

from typing import Any, Dict


def read_file(file_path: str) -> str:
    if not os.path.exists(file_path):
        logging.error(f"File '{file_path}' not found")
        raise Exception(f"File '{file_path}' not found")

    try:
        with open(file_path, "r", encoding="utf-8") as template_file:
            content = template_file.read()
        return content
    except Exception as e:
        logging.error(f"Read file '{file_path}' failed")
        raise

def format_file_with_args(file_content:str, args: Dict[str, Any])-> str:
    from io import StringIO

    formated = file_content
    original_stdout = sys.stdout
    buffer = StringIO()
    sys.stdout = buffer
    try:
        code = f"print(f\"\"\"{file_content}\"\"\")"
        exec(code, args)
        formated = buffer.getvalue()
    finally:
        sys.stdout = original_stdout

    return formated



import json
import datetime

from typing import Optional, List

from xgae.engine.engine_base import XGAToolSchema
from xgae.utils.misc import read_file, format_file_with_args


class XGAPromptBuilder():
    def __init__(self, system_prompt: Optional[str] = None):
        self.system_prompt = system_prompt
        self.prompt_builder_type = "system" if system_prompt else "general"


    def build_task_prompt(self,
                          model_name: str,
                          system_tool_schemas: List[XGAToolSchema],
                          general_tool_schemas: List[XGAToolSchema],
                          custom_tool_schemas: List[XGAToolSchema],
                          agent_tool_schemas: List[XGAToolSchema])-> str:
        if self.system_prompt is None:
            if general_tool_schemas and len(general_tool_schemas) > 0:
                self.system_prompt = self._load_general_prompt(model_name)
            else:
                system_prompt = read_file("templates/system_prompt_template.md")
                self.system_prompt = "#System Prompt" + "\n" + system_prompt
                self.prompt_builder_type = "system"
            task_prompt = self.system_prompt
        else:
            system_tool_prompt = read_file("templates/system_prompt_template.md")
            task_prompt = self.system_prompt + "\n" + system_tool_prompt

        if self.prompt_builder_type == "system":
            system_tool_prompt = ""
            general_tool_prompt = self.build_mcp_tool_prompt(general_tool_schemas)
            general_example_prompt = ""
        else:
            system_tool_prompt, _= self.build_openai_tool_prompt(system_tool_schemas)
            general_tool_prompt, general_example_prompt = self.build_openai_tool_prompt(general_tool_schemas)

        custom_tool_prompt = self.build_mcp_tool_prompt(custom_tool_schemas)
        agent_tool_prompt = self.build_mcp_tool_prompt(agent_tool_schemas)

        if self.prompt_builder_type == "system":
            task_prompt = task_prompt.replace("<<general_tool_schemas>>", general_tool_prompt)
            task_prompt = task_prompt.replace("<<custom_tool_schemas>>", custom_tool_prompt)
            task_prompt = task_prompt.replace("<<agent_tool_schemas>>", agent_tool_prompt)
        else:
            task_prompt = task_prompt.replace("<<system_tool_schemas>>", system_tool_prompt)
            task_prompt = task_prompt.replace("<<general_tool_schemas>>", general_tool_prompt)
            task_prompt = task_prompt.replace("<<general_tool_examples>>", general_example_prompt)
            task_prompt = task_prompt.replace("<<custom_tool_schemas>>", custom_tool_prompt)
            task_prompt = task_prompt.replace("<<agent_tool_schemas>>", agent_tool_prompt)

        return task_prompt


    def build_openai_tool_prompt(self, tool_schemas:List[XGAToolSchema]):
        tool_schemas = tool_schemas or []
        example_prompt = ""
        schema_prompt = ""

        if len(tool_schemas) > 0:
            example_prompt = ""
            openai_schemas = []
            for tool_schema in tool_schemas:
                openai_schema = {}
                openai_function = {}
                openai_schema['type'] = "function"
                openai_schema['function'] = openai_function

                openai_function['name'] = tool_schema.tool_name
                openai_function['description'] = tool_schema.description if tool_schema.description else 'No description available'
                openai_parameters = {}
                openai_function['parameters'] = openai_parameters

                input_schema = tool_schema.input_schema
                openai_parameters['type']       = input_schema['type']
                openai_parameters['properties'] = input_schema.get('properties', {})
                openai_parameters['required']   = input_schema['required']

                openai_schemas.append(openai_schema)

                metadata = tool_schema.metadata or {}
                example = metadata.get("example", None)
                if example:
                    example_prompt += f"\n{example}\n"

            schema_prompt = json.dumps(openai_schemas, ensure_ascii=False, indent=2)

        return schema_prompt, example_prompt


    def build_mcp_tool_prompt(self, tool_schemas:List[XGAToolSchema])-> str:
        tool_schemas = tool_schemas or []
        schema_prompt = ""

        if len(tool_schemas) > 0:
            for tool_schema in tool_schemas:
                description = tool_schema.description if tool_schema.description else 'No description available'
                schema_prompt += f"- {tool_schema.tool_name}: {description}\n"
                parameters = tool_schema.input_schema.get('properties', {})
                schema_prompt += f"   Parameters: {parameters}\n"
                schema_prompt += "\n"

        return schema_prompt


    def _load_general_prompt(self, model_name) -> Optional[str]:
        # TODO: Future
        # if "gemini-2.5-flash" in model_name.lower() and "gemini-2.5-pro" not in model_name.lower():
        #     system_prompt_template = read_file("templates/gemini_system_prompt_template.txt")
        # else:
        #     system_prompt_template = read_file("templates/general_prompt_template.md")

        system_prompt_template = read_file("templates/general_prompt_template.md")
        system_prompt = format_file_with_args(system_prompt_template, {"datetime": datetime})

        system_prompt = system_prompt.format(
            current_date=datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d'),
            current_time=datetime.datetime.now(datetime.timezone.utc).strftime('%H:%M:%S'),
            current_year=datetime.datetime.now(datetime.timezone.utc).strftime('%Y')
        )

        # TODO: Future
        # if "anthropic" in model_name.lower():
        #     sample_response = read_file("templates/anthropic_response_sample.txt")
        #     system_prompt = system_prompt + "\n\n <sample_assistant_response>" + sample_response + "</sample_assistant_response>"

        return system_prompt


if __name__ == "__main__":
    import asyncio
    from xgae.engine.mcp_tool_box import XGAMcpToolBox
    from xgae.utils.setup_env import setup_logging

    setup_logging()

    async def main():
        ## Before Run Exec: uv run example-fault-tools
        mcp_tool_box = XGAMcpToolBox(custom_mcp_server_file="xga_mcp_servers.json")
        task_id = "task1"
        await mcp_tool_box.init_tool_schemas()
        await mcp_tool_box.creat_task_tool_box(task_id=task_id, general_tools=["*"], custom_tools=["*"])

        system_tool_schemas = mcp_tool_box.get_task_tool_schemas(task_id, "system")
        general_tool_schemas = mcp_tool_box.get_task_tool_schemas(task_id, "general")
        custom_tool_schemas = mcp_tool_box.get_task_tool_schemas(task_id, "custom")
        agent_tool_schemas = mcp_tool_box.get_task_tool_schemas(task_id, "agent")

        #system_prompt = read_file("templates/example/fault_user_prompt.md")
        system_prompt = None

        prompt_builder = XGAPromptBuilder(system_prompt)
        task_prompt = prompt_builder.build_task_prompt(model_name="openai/qwen-plus",
                                         system_tool_schemas=system_tool_schemas,
                                         general_tool_schemas=general_tool_schemas,
                                         custom_tool_schemas=custom_tool_schemas,
                                         agent_tool_schemas=agent_tool_schemas)

        print(task_prompt)


    asyncio.run(main())


import asyncio

from xgae.engine.mcp_tool_box import XGAMcpToolBox
from xgae.engine.task_engine import XGATaskEngine

from xgae.utils.misc import read_file

from xgae.utils.setup_env import setup_langfuse, setup_logging

setup_logging()
langfuse = setup_langfuse()

async def main() -> None:
    # Before Run Exec: uv run example-fault-tools
    tool_box = XGAMcpToolBox(custom_mcp_server_file="xga_mcp_servers.json")
    system_prompt = read_file("templates/example/fault_user_prompt.md")

    engine = XGATaskEngine(tool_box=tool_box,
                           general_tools=[],
                           custom_tools=["*"],
                           system_prompt=system_prompt)

    # Two task run in same langfuse trace
    trace_id = langfuse.trace(name="xgae_example_run_human_in_loop").trace_id

    user_input =  "locate fault and solution"
    final_result = await engine.run_task_with_final_answer(
        task_input={"role": "user", "content": user_input},
        trace_id=trace_id
    )
    print("FINAL RESULT:", final_result)

    if final_result["type"] == "ask":
        print("====== Wait for user input ... ======")
        user_input = "ip=10.0.1.1"
        final_result = await engine.run_task_with_final_answer(
            task_input={"role": "user", "content": user_input},
            trace_id=trace_id
        )
        print("FINAL RESULT:", final_result)

asyncio.run(main())


# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the terms described in the LICENSE file in
# the root directory of this source tree.


import datetime
import json
import logging
import os
from typing import Any

from dotenv import load_dotenv
from are.simulation.apps.agent_user_interface import AgentUserInterface
from are.simulation.apps.mcp.mcp_app import MCPApp
from are.simulation.scenarios.scenario import Scenario, ScenarioValidationResult
from are.simulation.scenarios.utils.env_utils import expand_env_vars

from are.simulation.scenarios.utils.registry import register_scenario
from are.simulation.types import EventRegisterer, event_registered

# Set up logger
logger = logging.getLogger(__name__)

# Load environment variables from .env file
load_dotenv()



@register_scenario("scenario_bomc_fault")
class ScenarioBomcFault(Scenario):
    start_time: float | None = datetime.datetime.now().timestamp()
    duration: float | None = None

    def init_and_populate_apps(self, *args, **kwargs) -> None:

        # get apps from the demo
        agui = AgentUserInterface()
        self.apps = [agui]

        # form task from universe params
        self.start_time = datetime.datetime.now().timestamp()

        # Load additional MCP apps from JSON file if specified
        self._load_mcp_apps_from_json()


    def build_events_flow(self) -> None:
        """Define the sequence of events that will occur during the scenario"""

        agui = self.get_typed_app(AgentUserInterface)

        with EventRegisterer.capture_mode():
            # User event: User requests task creation
            event1 = agui.send_message_to_agent(
                content="locate 10.0.0.1 fault and solution",
            ).depends_on(None, delay_seconds=2)

            oracle1 = (
                agui.send_message_to_user(
                    content="Fault Cause: Phone Recharge Application Crash; Solution: Restart Phone Recharge Application",
                )
                .oracle()
                .depends_on(event1, delay_seconds=10)
            )


        self.events = [event1]


    def validate(self, env) -> ScenarioValidationResult:
        """
        Validate that the scenario completed successfully.

        Check that the agent properly interacted with our custom app.
        """
        try:
            aui = env.get_app("AgentUserInterface")
            msg = aui.get_last_message_from_agent()
            content = msg.content
            if "Phone Recharge Application Crash" in content:
                return ScenarioValidationResult(success=True)
            else:
                return ScenarioValidationResult(success=False)
        except Exception as e:
            return ScenarioValidationResult(success=False, exception=e)


    def _load_mcp_apps_from_json(self) -> None:
        """
        Load additional MCP apps from a JSON file specified in the environment variables.

        The JSON file should follow the Claude MCP definition format with a ``mcpServers``
        key containing server configurations. Environment variables in the JSON are
        expanded using the current environment.

        :return: None
        """
        # Get the JSON file path from environment variables
        mcp_apps_json_path = os.environ.get("MCP_APPS_JSON_PATH")

        if not mcp_apps_json_path:
            return

        # Ensure self.apps is initialized
        if not hasattr(self, "apps") or self.apps is None:
            self.apps = []

        try:
            with open(mcp_apps_json_path, "r") as f:
                mcp_config = json.load(f)

            if "mcpServers" not in mcp_config:
                logger.warning(
                    f"No 'mcpServers' key found in MCP apps JSON file: {mcp_apps_json_path}"
                )
                return

            # Create MCP apps from the JSON configuration
            for server_name, server_config in mcp_config["mcpServers"].items():
                mcp_app = self._create_mcp_app(server_name, server_config)
                if mcp_app:
                    self.apps.append(mcp_app)
                    logger.info(f"Added MCP app: {server_name}")

        except Exception as e:
            logger.error(f"Error loading MCP apps from JSON: {e}", exc_info=True)

    def _create_mcp_app(self, name: str, config: dict[str, Any]) -> MCPApp | None:
        """
        Create an MCPApp instance from a server configuration.

        Supports two types of MCP servers:

        1. Local servers with command and arguments
        2. Remote SSE servers with URL and headers

        Environment variables in the configuration are expanded before creating the app.

        :param name: The name of the MCP app
        :type name: str
        :param config: The server configuration from the JSON file
        :type config: dict[str, Any]
        :return: An MCPApp instance or None if creation fails
        :rtype: MCPApp | None
        """
        try:
            # Expand environment variables in the configuration
            expanded_config = expand_env_vars(config, allowed=["HF_TOKEN"])

            # Handle SSE remote servers
            if expanded_config.get("type") == "sse" or "url" in expanded_config:
                return MCPApp(
                    name=name,
                    server_url=expanded_config.get("url"),
                    sse_headers=expanded_config.get("headers", {}),
                )

            # Handle local servers with command and args
            elif "command" in expanded_config:
                return MCPApp(
                    name=name,
                    server_command=expanded_config.get("command"),
                    server_args=expanded_config.get("args", []),
                    server_env=expanded_config.get("env", {}),
                )

            else:
                logger.warning(f"Unsupported MCP server configuration for {name}")
                return None

        except Exception as e:
            logger.error(f"Error creating MCP app {name}: {e}", exc_info=True)
            return None


# Before Run Scenario, should xage project , execute command:
# 1. Vailid Success: uv run example-fault-tools
# 2. Vailid Fail: uv run example-fault-tools --alarmtype 2
if __name__ == "__main__":
    from are.simulation.scenarios.utils.cli_utils import run_and_validate

    run_and_validate(ScenarioBomcFault())



import logging

from xgae.utils.setup_env import setup_langfuse, setup_logging

setup_logging()
langfuse = setup_langfuse()

langfuse.span()
trace_id = langfuse.trace(name="xgae_test_trace").trace_id
logging.info(f"1 trace_id={trace_id}")
trace = langfuse.trace(id=trace_id)
logging.info(f"2 trace_id={trace.trace_id}")

trace.update(input={"task_id": "trace_input"})
trace.event(name="trace_event1", input="trace event1 input", output="trace event1 output")

span1 = trace.span(name="x_span_1", input={"task_id": "span1"})
logging.info(f"span1_name={span1}")
span1.event(name="span1_event", input="span1 event input", output="span1 event output")
span1.end(output={"result": "span1_result"})


span1 = langfuse.span(trace_id=trace.id, id=span1.id)
span2_id = span1.span(name="x_span_2", input={"task_id": "span2"}).id
logging.info(f"span2_id={span2_id}")
span2 = trace.span(id=span2_id)
span2.event(name="span2_event", input="span1 event input", output="span2 event output")
span2.end(output={"result": "span2_result"})

trace.event(name="trace_event2", input="trace event2 input", output="trace event2 output")
trace.update(output={"result": "trace_output"})


import logging
from typing import Any, Optional, Callable, Literal, Dict, List
from typing_extensions import override

from xgae.engine.engine_base import XGAToolBox
from xgae.engine.task_engine import XGATaskEngine
from xgae.engine.prompt_builder import XGAPromptBuilder
from xgae.utils.llm_client import LLMConfig


class ARETaskEngine(XGATaskEngine):
    def __init__(self,
                 agent: Any,
                 agent_id: str,
                 max_auto_run: int,
                 llm_config: Optional[LLMConfig] = None,
                 tool_box: Optional[XGAToolBox] = None,
                 prompt_builder: Optional[XGAPromptBuilder] = None,
                 pre_run_task_fn : Callable[[Any, int, List[Dict[str, Any]]], Any] = None,
                 post_run_task_fn : Callable[[Any, int, Dict[str, Any]], Any] = None,
                 terminate_task_fn : Callable[[Any, int], bool] = None,
                ):
        super().__init__(agent_id       = agent_id,
                         general_tools  = [],
                         custom_tools   = ["*"],
                         max_auto_run   = max_auto_run,
                         llm_config     = llm_config,
                         tool_box       = tool_box,
                         prompt_builder = prompt_builder,
                         )
        self.agent = agent
        self.pre_run_task_fn = pre_run_task_fn
        self.post_run_task_fn = post_run_task_fn
        self.terminate_task_fn = terminate_task_fn


    @override
    def running_task_checkpoint(self,
                                task_state: Literal["before_completion", "after_completion", "termination_check"],
                                iterations: int,
                                llm_messages: List[Dict[str, Any]] = None,
                                llm_response: Dict[str, Any] = None
                                )-> bool:
        is_continue_task = True

        if task_state == "before_completion" and self.pre_run_task_fn:
            self.pre_run_task_fn(self.agent, iterations, llm_messages)
        elif task_state == "after_completion" and self.post_run_task_fn:
            self.post_run_task_fn(self.agent, iterations, llm_response)
        elif task_state == "termination_check":
            if self.terminate_task:
                logging.warning(f"running_task_checkpoint: ⚠️ TASK: {self.task_id} STOP RUNNING for STOP Command !")

            if self.terminate_task_fn:
                is_terminate = self.terminate_task_fn(self.agent, iterations) if self.terminate_task_fn  else False
                if is_terminate:
                    logging.warning(f"running_task_checkpoint: ⚠️ TASK: {self.task_id} STOP RUNNING for Termination Function !")
                    self.stop_task()

            is_continue_task = not self.terminate_task

        return is_continue_task



if __name__ == "__main__":
    import asyncio
    import os
    from xgae.utils.misc import read_file
    from xgae.utils.setup_env import setup_logging
    from xgae.engine.mcp_tool_box import XGAMcpToolBox
    from xgae.engine.prompt_builder import XGAPromptBuilder

    setup_logging()

    def pre_run_task(agent, iterations:int, llm_messages: List[Dict[str, Any]]):
        prompt = "\n\n".join([f"{key}: {value}" for d in llm_messages for key, value in d.items()]) if llm_messages else ""
        logging.info(f"pre_run_task: iterations={iterations}, prompt: \n{prompt}\n")


    def post_run_task(agent, iterations: int, llm_response: Dict[str, Any]):
        logging.info(f"post_run_task: iterations={iterations}, prompt: \n{llm_response}\n")


    def terminate_task(agent, iterations: int) -> bool:
        logging.info(f"terminate_task: iterations={iterations}")
        return iterations > 8 # can test terminate by > 3


    async def main():
        # Before Run Exec: uv run example-fault-tools
        # LLAMA_API_KEY , LLAMA_API_BASE
        tool_box = XGAMcpToolBox(custom_mcp_server_file="xga_mcp_servers.json")
        system_prompt = read_file("templates/example/fault_user_prompt.md")
        prompt_builder = XGAPromptBuilder(system_prompt)
        llm_config = LLMConfig(
            model            = "openai/qwen-plus",
            api_key          = os.getenv('LLAMA_API_KEY') ,
            api_base         = os.getenv('LLAMA_API_BASE'),
            stream           =True,
            enable_thinking  = False,
        )

        engine =  ARETaskEngine(
            agent               = "AREAgent", # Just for test,ARE use real Agent Object
            agent_id            = "agent_1",
            max_auto_run        = 15,
            llm_config          = llm_config,
            tool_box            = tool_box,
            prompt_builder      = prompt_builder,
            pre_run_task_fn     = pre_run_task,
            post_run_task_fn    = post_run_task,
            terminate_task_fn   = terminate_task
        )

        user_input = "locate 10.0.0.1 fault and solution"
        chunks = []
        async for chunk in engine.run_task(task_input={"role": "user", "content": user_input}):
            chunks.append(chunk)
            print(chunk)

        final_result = engine.parse_final_result(chunks)
        print(f"\n\nFINAL_RESULT: {final_result}")


    asyncio.run(main())


import logging

from typing import List, Optional, Dict, Any

from xgae.engine.engine_base import XGAError, XGAToolSchema, XGAToolBox, XGAToolResult, XGAToolType

class XGASystemTools:
    SYSTEM_SERVER_NAME = "XgaSystem"

    def get_sys_tool_schemas(self) -> List[XGAToolSchema] :
        sys_tool_schemas = []
        sys_tool_schemas.append(self._get_complete_tool_schema())
        sys_tool_schemas.append(self._get_ask_tool_schema())
        return sys_tool_schemas


    def call_tool(self, tool_name: str, args: Optional[Dict[str, Any]] = None) -> XGAToolResult:
        args = args or {}
        try:
            method = getattr(self, tool_name)
        except AttributeError:
            raise ValueError(f"XGASystemTools call_tool: System Tool function '{tool_name}' not found.")

        if not callable(method):
            raise ValueError(f"XGASystemTools call_tool: '{tool_name}' is not a callable function.")

        return method(**args)


    def complete(self, task_id: str, text: str=None, attachments: str=None)->XGAToolResult :
        logging.info(f"<XGAETools-complete>: task_id={task_id}, text={text}, attachments={attachments}")
        return XGAToolResult(success=True, output=str({"status": "complete"}))


    def ask(self, task_id: str, text: str=None, attachments: str=None)->XGAToolResult :
        logging.info(f"<XGAETools-ask>: task_id={task_id}, text={text}, attachments={attachments}")
        return XGAToolResult(success=True, output=str({"status": "Awaiting user response..."}))


    def _get_complete_tool_schema(self)->XGAToolSchema :
        return XGAToolSchema(
            tool_name       = "complete",
            function_name   = "complete",
            tool_type       = "system",
            server_name     = self.SYSTEM_SERVER_NAME,
            description     = """A special tool to indicate you have completed all tasks and are about to enter complete state. Use ONLY when: 1) All tasks in todo.md are marked complete [x], 2) The user's original request has been fully addressed, 3) There are no pending actions or follow-ups required, 4) You've delivered all final outputs and results to the user. IMPORTANT: This is the ONLY way to properly terminate execution. Never use this tool unless ALL tasks are complete and verified. Always ensure you've provided all necessary outputs and references before using this tool. Include relevant attachments when the completion relates to specific files or resources.""",
            input_schema    = {
                'type': 'object',
                'required': ['text'],
                'properties': {
                    'text': {
                        'type': 'string',
                        'default': None,
                        'description': 'Completion summary. Include: 1) Task summary 2) Key deliverables 3) Next steps 4) Impact achieved'
                    },
                    'attachments': {
                        'anyOf': [{'type': 'string'}, {'type': 'null'}],
                        'default': None,
                        'description': 'Comma-separated list of final outputs. Use when: 1) Completion relates to files 2) User needs to review outputs 3) Deliverables in files'
                    }
                }
            },
            metadata        = None
        )


    def _get_ask_tool_schema(self)->XGAToolSchema :
        return XGAToolSchema(
            tool_name       = "ask",
            function_name   = "ask",
            tool_type       = "system",
            server_name     = self.SYSTEM_SERVER_NAME,
            description     = """Ask user a question and wait for response. Use for: 1) Requesting clarification on ambiguous requirements, 2) Seeking confirmation before proceeding with high-impact changes, 3) Gathering additional information needed to complete a task, 4) Offering options and requesting user preference, 5) Validating assumptions when critical to task success, 6) When encountering unclear or ambiguous results during task execution, 7) When tool results don't match expectations, 8) For natural conversation and follow-up questions, 9) When research reveals multiple entities with the same name, 10) When user requirements are unclear or could be interpreted differently. IMPORTANT: Use this tool when user input is essential to proceed. Always provide clear context and options when applicable. Use natural, conversational language that feels like talking with a helpful friend. Include relevant attachments when the question relates to specific files or resources. CRITICAL: When you discover ambiguity (like multiple people with the same name), immediately stop and ask for clarification rather than making assumptions.""",
            input_schema    = {
                'type': 'object',
                'required': ['text'],
                'properties': {
                    'text': {
                        'type': 'string',
                        'default': None,
                        'description': 'Question text to present to user. Include: 1) Clear question/request 2) Context why input is needed 3) Available options 4) Impact of choices 5) Relevant constraints'
                    },
                    'attachments': {
                        'anyOf': [{'type': 'string'}, {'type': 'null'}],
                        'default': None,
                        'description': 'Comma-separated list of files/URLs to attach. Use when: 1) Question relates to files/configs 2) User needs to review content 3) Options documented in files 4) Supporting evidence needed'
                    }
                }
            },
            metadata        = None
        )


if __name__ == "__main__":
    from xgae.utils.setup_env import setup_logging
    setup_logging()

    system_tools = XGASystemTools()

    tool_name = "complete"
    args = {
        'task_id': "task1",
        'text': "complete task",
        'attachments': 'complete attachments'
    }
    result = system_tools.call_tool(tool_name, args)
    print(result)

    tool_name = "ask"
    args = {
        'task_id': "task2",
        'text': "ask task",
        'attachments': 'ask attachments'
    }
    result = system_tools.call_tool(tool_name, args)
    print(result)


import asyncio

from xgae.engine.mcp_tool_box import XGAMcpToolBox
from xgae.engine.task_engine import XGATaskEngine
from xgae.utils.misc import read_file

from xgae.utils.setup_env import setup_logging


is_stream = False
if is_stream:
    setup_logging(log_level="ERROR") # only show chunk
else:
    setup_logging()

system_prompt = read_file("templates/example/fault_user_prompt.md")
tool_box = XGAMcpToolBox(custom_mcp_server_file="xga_mcp_servers.json")

# Before Run Exec: uv run example-fault-tools --alarmtype 2  , uv run example-a2a-tools
# If want to use real A2A agent tool, use 'xgaproxy' project run: v run xga-a2a-proxy
async def run_custom_plus_a2a_tools() -> None:
    engine = XGATaskEngine(tool_box=tool_box,
                           custom_tools=["*"],
                           system_prompt=system_prompt
                           )


    user_input =  "locate 10.2.3.4 fault and solution"
    global is_stream
    if is_stream:
        chunks = []
        async for chunk in engine.run_task(task_input={"role": "user", "content": user_input}):
            chunks.append(chunk)
            print(chunk)

        final_result = engine.parse_final_result(chunks)
        print(f"\n\nFINAL_RESULT: {final_result}")
    else:
        final_result = await engine.run_task_with_final_answer(task_input={"role": "user", "content": user_input})
        print(f"\n\nFINAL_RESULT: {final_result}")


# Before run example execute with 'xgatools' project: uv run xgatools
# Before run exec: uv run example-fault-tools --alarmtype 3  , uv run example-a2a-tools
async def run_custom_plus_general_tools() -> None:
    engine = XGATaskEngine(system_prompt=system_prompt,
                           tool_box=tool_box,
                           general_tools=["web_search"],
                           custom_tools=["*"])

    user_input = "locate 10.3.4.5 fault and solution"
    final_result = await engine.run_task_with_final_answer(task_input={"role": "user", "content": user_input})
    print("FINAL RESULT:", final_result)


async def main() -> None:
    #await run_custom_plus_a2a_tools()
    await run_custom_plus_general_tools()

asyncio.run(main())


import asyncio

from uuid import uuid4

from xgae.utils.setup_env import setup_logging

from examples.agent.langgraph.reflection.reflection_agent import ReflectiontAgent, AgentContext


async def main():
    is_stream = True  # two mode agent experience
    task_no = 0
    user_inputs = [
        #"5+5", # For no tool call
        #"locate 10.2.3.4 fault and solution",  # For custom tool
        "locate fault and solution",   # For human append input
    ]

    for user_input in user_inputs:
        agent = ReflectiontAgent(use_prompt_rag=True)
        task_no += 1
        context: AgentContext = {
            'task_id': f"agent_task_{uuid4()}",   # can be set with request_id, must be unique
            'user_id': "agent_user_1",
            'agent_id': "agent_1",
        }

        is_resume = False
        auto_continue = True
        while auto_continue:
            if is_stream:
                print(f"*** START AGENT : RUN generate USER_INPUT={user_input}")
                async for chunk in agent.generate(user_input, context, is_resume):
                    type = chunk['type']
                    if type == "error" or type == "answer":
                        await asyncio.sleep(1)
                        print(f"FINAL_RESULT: {chunk}")
                        auto_continue = False
                    elif type == "ask":
                        print(f"ASK_USER: {chunk}")
                        user_input = "17.0.0.1"
                        is_resume = True
                        auto_continue = True
                    else:
                        print(f"RESULT_CHUNK: {chunk}")
                        auto_continue = False
            else:
                print(f"*** START AGENT : RUN generate_with_result  USER_INPUT={user_input}")
                result = await agent.generate_with_result(user_input, context, is_resume)
                await asyncio.sleep(1)
                type = result['type']
                if type == "error" or type == "answer":
                    print(f"FINAL_RESULT: {result}")
                    auto_continue = False
                elif type == "ask":
                    print(f"ASK_USER: {result}")
                    user_input = "18.0.0.1"
                    is_resume = True
                    auto_continue = True


if __name__ == "__main__":
    setup_logging()
    asyncio.run(main())


from typing import Any, Dict, List, TypedDict, Optional

from xgae.engine.engine_base import XGATaskResult

class EvaluateResult(TypedDict, total=False):
    task_result: Dict[str, Any]
    task_process: Dict[str, Any]
    function_call: Dict[str, Any]

class AgentContext(TypedDict, total=False):
    task_id: str
    session_id: str
    user_id: str
    agent_id: str
    thread_id: str


class TaskState(TypedDict, total=False):
    """State definition for the agent orchestration graph"""
    llm_messages: List[Dict[str, Any]]
    user_inputs: List[str]
    next_node: str
    system_prompt: str
    custom_tools: List[str]
    general_tools: List[str]
    task_result: XGATaskResult
    final_result: XGATaskResult
    eval_result: EvaluateResult
    retry_count: int
    task_no: int
    agent_context: AgentContext





import re
from typing import Optional, List, Literal

from xgae.engine.prompt_builder import XGAPromptBuilder
from xgae.utils.misc import read_file


class XGAArePromptBuilder(XGAPromptBuilder):
    def __init__(self, general_system_prompt: Optional[str]):
        are_system_prompt = self.build_are_system_prompt(general_system_prompt)
        super().__init__(are_system_prompt)


    def build_are_system_prompt(self, _system_prompt: str)-> str:
        pattern = r'<general_instructions>(.*?)</general_instructions>'
        prompt_are_general = re.search(pattern, _system_prompt, re.DOTALL)
        prompt_header = "# CORE IDENTITY\n"
        if prompt_are_general:
            prompt_are_general = prompt_header + prompt_are_general.group(1).strip() + "\n\n"
        else:
            prompt_are_general = prompt_header + _system_prompt  + "\n\n"

        are_tool_prompt = read_file("templates/are_prompt_template.md")
        are_system_prompt = prompt_are_general + are_tool_prompt

        return are_system_prompt




import asyncio
import json
import threading
import time

from concurrent.futures import Future as ConcurrentFuture
from typing import Any, Optional, Coroutine, List, Dict

from mcp.client.session import ClientSession
from mcp.client.sse import sse_client

from xgae.engine.engine_base import  XGAToolResult


class AsyncToolExecutor:
    _instance = None
    _loop = None
    _thread = None
    _lock = threading.Lock()

    def __new__(cls):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super().__new__(cls)
                cls._init_event_loop()
            return cls._instance


    @classmethod
    def _init_event_loop(cls):
        if cls._loop is None or cls._loop.is_closed():
            cls._loop = asyncio.new_event_loop()

        def start_loop(loop):
            asyncio.set_event_loop(loop)
            try:
                loop.run_forever()
            finally:
                if not loop.is_closed():
                    loop.close()

        cls._thread = threading.Thread(target=start_loop, args=(cls._loop,), daemon=True)
        cls._thread.start()

        time.sleep(0.1)

    @classmethod
    def submit(cls, coro: Coroutine) -> ConcurrentFuture:
        if cls._loop is None or cls._loop.is_closed():
            cls._init_event_loop()

        return asyncio.run_coroutine_threadsafe(coro, cls._loop)

    @classmethod
    def run_and_wait(cls, coro: Coroutine, timeout: Optional[float] = None) -> Any:
        future = cls.submit(coro)
        return future.result(timeout=timeout)

    @classmethod
    def get_loop(cls):
        if cls._loop is None or cls._loop.is_closed():
            cls._init_event_loop()
        return cls._loop

    @classmethod
    def shutdown(cls):
        if cls._loop is not None and not cls._loop.is_closed():
            cls._loop.call_soon_threadsafe(cls._loop.stop)

from mcp.types import CallToolResult

async def _mcp_sse_tool_call(url: str, tool_name: str, arguments: dict[str, Any]) -> CallToolResult:
    async with sse_client(url, sse_read_timeout=5) as streams:
        async with ClientSession(*streams) as session:
            await session.initialize()
            return  await session.call_tool(tool_name, arguments)


def call_mcp_tool(url: str, tool_name: str, arguments: dict[str, Any], timeout: Optional[float] = None) -> Any:
    future = AsyncToolExecutor.submit(_mcp_sse_tool_call(url, tool_name, arguments))
    mcp_result:CallToolResult =  future.result(timeout=timeout)
    result = {
        'isError': mcp_result.isError,
    }
    if mcp_result.isError:
        result['content'] = mcp_result.content[0].text
    else:
        result['content'] = mcp_result.structuredContent['result']

    return  json.dumps(result)


def convert_mcp_tool_result(org_result: str)->XGAToolResult:
    result = XGAToolResult(success=True, output=str(org_result))

    if org_result and isinstance(org_result, str) and "isError" in org_result:
        try:
            _result:dict = json.loads(org_result)
            content = _result.get('content', None)
            isError = _result.get('isError', None)

            if content and isError:
                success = not bool(isError)
                result = XGAToolResult(success=bool(success), output=str(content))
        except:
            pass

    return result


if __name__ == "__main__":
    async def sample_task(name, duration):
        print(f"Task '{name}' Begin，time = {duration} ")
        await asyncio.sleep(duration)
        return f"Task '{name}' Finished"

    async def main():
        try:
            # result = AsyncToolExecutor.run_and_wait(sample_task("mytool", 1))
            # print(f"Result1: {result}")

            result = call_mcp_tool("http://localhost:17070/sse", "get_alarm", {"ip":"13.0.0.13"})
            result = convert_mcp_tool_result(result)
            print(f"Result2: {result}")

            result = call_mcp_tool("http://localhost:17070/sse", "get_alarm1", {"ip":"13.0.0.13"})
            result = convert_mcp_tool_result(result)
            print(f"Result3: {result}")

        except Exception as e:
            print(f"Thread Fail: {e}")

    with asyncio.Runner() as runner:
            runner.run(main())


import asyncio
import sys

from xgae.engine.mcp_tool_box import XGAMcpToolBox
from xgae.engine.task_engine import XGATaskEngine

from xgae.utils.misc import read_file

from xgae.utils.setup_env import setup_langfuse, setup_env_logging

setup_env_logging()
langfuse = setup_langfuse()

def get_user_message(question)-> str:
    while True:
        user_message = input(f"\n💬 {question}: ")
        if user_message.lower() == 'exit' or user_message.lower() == 'quit':
            print("\n====== Extreme General Agent Engine CLI EXIT ======")
            sys.exit()

        if not user_message.strip():
            print("\nuser message is empty， input agin ！！！\n")
            continue

        return user_message

async def cli() -> None:
    await asyncio.sleep(1)
    print("\n====== Extreme General Agent Engine CLI START ======")
    user_message = input("\n💬 Start Custom MCP Server and Load User Prompt (Yes/No): ")
    tool_box = XGAMcpToolBox(custom_mcp_server_file="xga_mcp_servers.json")
    system_prompt = None
    general_tools = []
    custom_tools = []
    if user_message.lower() == 'yes':
        print(f"--- Start Custom MCP Server in custom_servers.json")
        print(f"--- Load User Prompt in example/fault_user_prompt.md")
        system_prompt = read_file("templates/example/fault_user_prompt.md")
        custom_tools = ["*"]
    else:
        print(f"--- Start General Agent Server")
        print(f"--- Load System Prompt")
        general_tools = ["*"]

    while True:
        user_message = get_user_message("Enter your task input message (or 'exit' to quit)")

        print("\n🔄 Running XGA Engine ...\n")
        engine = XGATaskEngine(tool_box=tool_box,
                               general_tools=general_tools,
                               custom_tools=custom_tools,
                               system_prompt=system_prompt)

        # Two task run in same langfuse trace
        trace_id = langfuse.trace(name="xgae_cli").trace_id
        auto_continue = True
        while auto_continue:
            auto_continue = False
            final_result = await engine.run_task_with_final_answer(
                task_input={'role': "user", 'content': user_message},
                trace_id=trace_id
            )

            if final_result["type"] == "ask":
                await asyncio.sleep(1)
                print(f"\n📌 ASK INFO: {final_result['content']}")
                user_message = get_user_message("Enter ASK information (or 'exit' to quit)")
                auto_continue = True
                continue

            await asyncio.sleep(1)
            result_prefix = "✅" if final_result["type"] == "answer" else "❌"
            print(f"\n {result_prefix} FINAL RESULT: {final_result['content']}")


def main():
    asyncio.run(cli())


if __name__ == "__main__":
    main()


import logging
import os

from dotenv import load_dotenv
from langfuse import Langfuse

from xgae.utils import to_bool

load_dotenv()

def setup_logging(log_file: str=None, log_level: str="INFO") :
    import colorlog

    logging_level = getattr(logging, log_level.upper(), logging.INFO)

    logger = logging.getLogger()
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    log_colors = {
        'DEBUG': 'cyan',
        'INFO': 'green',
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'red,bg_white'
    }

    console_formatter = colorlog.ColoredFormatter('%(log_color)s%(asctime)s - %(levelname)-8s%(reset)s %(white)s%(message)s',
        log_colors=log_colors,
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)

    if log_file:
        log_dir = os.path.dirname(log_file)
        if log_dir and not os.path.exists(log_dir):
            os.makedirs(log_dir, exist_ok=True)
        else:
            os.remove(log_file)

        file_formatter = logging.Formatter(
            '%(asctime)s -%(levelname)-8s  %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        )

        file_handler = logging.FileHandler(log_file, encoding='utf-8')
        file_handler.setFormatter(file_formatter)
        logger.addHandler(file_handler)

    logger.setLevel(logging_level)

    logging.info(f"🛠️ XGA_LOGGING is initialized, log_level={log_level}, log_file={log_file}")

def setup_env_logging():
    log_enable = to_bool(os.getenv("LOG_ENABLE", True))
    log_level = os.getenv("LOG_LEVEL", "INFO")
    log_file = os.getenv("LOG_FILE", "log/xga.log")
    if log_enable :
        setup_logging(log_file, log_level)
        setup_logging(log_file, log_level)

def setup_langfuse() -> Langfuse:
    env_public_key = os.getenv("LANGFUSE_PUBLIC_KEY")
    env_secret_key = os.getenv("LANGFUSE_SECRET_KEY")
    env_host = os.getenv("LANGFUSE_HOST", "https://cloud.langfuse.com")
    if env_public_key and env_secret_key:
        _langfuse = Langfuse(enabled=True,
                            public_key=env_public_key,
                            secret_key=env_secret_key,
                            host=env_host)

        logging.info("🛠️ XGA_LANGFUSE initialized Successfully by Key !")
    else:
        _langfuse = Langfuse(enabled=False)
        logging.warning("🛠️ XGA_LANGFUSE Not set key, Langfuse is disabled!")

    return _langfuse

if __name__ == "__main__":
    langfuse = setup_langfuse()
    logging.warning(f"langfuse is enable={langfuse.enabled}")



import re
import asyncio
from typing import List, Dict, Any

from mammoth.results import success
from typing_extensions import override

from are.simulation.agents.are_simulation_agent import AgentStoppedException
from are.simulation.agents.llm.types import MMObservation
from are.simulation.agents.multimodal import Attachment, attachments_to_pil
from are.simulation.agents.agent_log import (
    LLMInputLog,
    LLMOutputThoughtActionLog,
    StepLog,
    StopLog,
    SystemPromptLog,
    ThoughtLog,
    ToolCallLog,
    FinalAnswerLog,
    ObservationLog,
    ErrorLog
)
from are.simulation.agents.default_agent.base_agent import (
    BaseAgent,
    RunningState,
    get_offset_from_time_config_mode
)

from xgae.utils.llm_client import LLMConfig
from xgae.gaia2.are_engine import ARETaskEngine
from are.simulation.agents.xga.are_tool_box import XGAAreToolBox
from are.simulation.agents.xga.are_prompt_builder import XGAArePromptBuilder

def pre_run_task_check(agent, iterations: int, llm_messages: List[Dict[str, Any]]):
    try:
        agent.logger.info(f"\n\n------  Starting Run Task Iteration {iterations}...   ------")

        if iterations == 0:
            agent.system_prompt = agent.task_engine.task_prompt
            agent.logger.info(f"------  SYSTEM_PROMPT   ------ \n{agent.system_prompt}\n")
            agent.append_agent_log(
                SystemPromptLog(
                    content     = agent.system_prompt,
                    timestamp   = agent.make_timestamp(),
                    agent_id    = agent.agent_id,
                )
            )

        agent.append_agent_log(
            StepLog(
                iteration   = agent.iterations,
                timestamp   = agent.make_timestamp(),
                agent_id    = agent.agent_id,
            )
        )

        if agent.stop_event.is_set():
            agent.logger.info(f"pre_run_task_check[{iterations}]: Recv Stop Event before condition, raise AgentStoppedException")
            raise AgentStoppedException("Agent stopped.")

        # Execute a pre_step() function if it exists
        for conditional_step in agent.conditional_pre_steps:
            if conditional_step.condition is None or conditional_step.condition(agent):
                conditional_step.function(agent)

        if agent.stop_event.is_set():
            agent.logger.info(f"pre_run_task_check[{iterations}]: Recv Stop Event after condition, raise AgentStoppedException")
            raise AgentStoppedException("Agent stopped.")

        # Begin step()
        agent.build_history_from_logs(
            exclude_log_types=["tool_call", "rationale", "action"]
        )

        agent.append_agent_log(
            LLMInputLog(
                content     = llm_messages,
                timestamp   = agent.make_timestamp(),
                agent_id    = agent.agent_id
            )
        )

        if agent.simulated_generation_time_config is not None:
            if agent.pause_env is None:
                raise ValueError("pause_env is not set")
            agent.pause_env()
    except Exception as e:
        agent.log_error(e)
        agent.logger.info(f"pre_run_task_check[{iterations}]: Exception Occur, Stop task")
        agent.task_engine.stop_task()


def post_run_task_check(agent, iterations: int, llm_response: Dict[str, Any]):
    try:
        agent.logger.info(f"------  LLM Response Iteration [{iterations}]   ------ \n{llm_response}\n")

        llm_output = str(llm_response)

        # Resume the environment after the generation of a thought/action if needed
        if agent.simulated_generation_time_config is not None:
            if agent.resume_env is None:
                raise ValueError("resume_env is not set")

            offset = get_offset_from_time_config_mode(
                time_config = agent.simulated_generation_time_config,
                completion_duration = 0,
            )
            agent.logger.info(f"post_run_task_check[{iterations}]: Resuming environment with {offset} offset")
            agent.resume_env(offset)

        metadata = {}
        agent.append_agent_log(
            LLMOutputThoughtActionLog(
                content             = llm_output,
                timestamp           = agent.make_timestamp(),
                agent_id            = agent.agent_id,
                prompt_tokens       = metadata.get("prompt_tokens", 0),
                completion_tokens   = metadata.get("completion_tokens", 0),
                total_tokens        = metadata.get("total_tokens", 0),
                reasoning_tokens    = metadata.get("reasoning_tokens", 0),
                completion_duration = metadata.get("completion_duration", 0),
            )
        )
        # end step()

        if agent.stop_event.is_set():
            agent.logger.info(f"post_run_task_check[{iterations}]: Recv Stop Event, raise AgentStoppedException")
            raise AgentStoppedException("Agent stopped.")

        # Execute a post_step() function if it exists (polling the Meta Agents Research Environments notifications for example)
        for conditional_step in agent.conditional_post_steps:
            if conditional_step.condition is None or conditional_step.condition(agent):
                conditional_step.function(agent)
    except Exception as e:
        agent.log_error(e)
        agent.logger.info(f"post_run_task_check[{iterations}]: Exception Occur, Stop task")
        agent.task_engine.stop_task()
    finally:
        if agent.simulated_generation_time_config and agent.resume_env:
            agent.resume_env(0.0)  # Resume without advancing time

        agent.iterations += 1
        agent.planning_counter += 1


def terminate_task_check(agent, iterations: int) -> bool:
    is_terminate_task = False

    try:
        if agent.termination_step and agent.termination_step.condition:
            is_terminate_task = agent.termination_step.condition(agent)
            if is_terminate_task:
                agent.logger.info(f"terminate_task_check[{iterations}]: termination_step.condition is True")
    except Exception as e:
        agent.log_error(e)

    return is_terminate_task


class XGAAreAgent(BaseAgent):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.task_engine: ARETaskEngine = None


    @override
    def initialize(self, attachments: list[Attachment] | None = None, **kwargs) -> None:
        self.logs = []
        self.iterations = 0
        self.planning_counter = 0

        tool_box = XGAAreToolBox(self.tools)

        general_system_prompt = "\n\n".join(prompt for prompt in self.init_system_prompts.values())
        prompt_builder = XGAArePromptBuilder(general_system_prompt)

        model_config = self.llm_engine.model_config
        llm_config = LLMConfig(
            model       = model_config.model_name,
            api_key     = model_config.api_key,
            api_base    = model_config.endpoint
        )

        self.task_engine =  ARETaskEngine(
            agent               = self,
            agent_id            = self.agent_id,
            max_auto_run        = self.max_iterations,
            llm_config          = llm_config,
            tool_box            = tool_box,
            prompt_builder      = prompt_builder,
            pre_run_task_fn     = pre_run_task_check,
            post_run_task_fn    = post_run_task_check,
            terminate_task_fn   = terminate_task_check,
        )

        # Reload the agent state if logs are provided
        start_logs = kwargs.pop("start_logs", [])
        if start_logs:
            self.replay(start_logs)

        # Include additional image PILs directly into state stack.
        if attachments:
            images = attachments_to_pil(attachments)
            self.action_executor.inject_state({f"image_{i}": image for i, image in enumerate(images)})
            self.logger.debug(f"XGAAreAgent initialize: Injecting images into states for {len(images)} images")
            self.logger.debug(f"XGAAreAgent initialize: New Keys {','.join(self.action_executor.state.keys())}")

        self.initialized = True


    @override
    def execute_agent_loop(self) -> str | None | MMObservation:
        with asyncio.Runner() as runner:
            runner.run(self.async_execute_agent_loop())

        # We have reached a termination condition, execute the termination method
        if self.termination_step.function is not None and not self.stop_event.is_set():
            return self.termination_step.function(self)


    async def async_execute_agent_loop(self) -> str | None | MMObservation:
        chunks = []
        async for chunk in self.task_engine.run_task(task_input={"role": "user", "content": self.task}):
            chunks.append(chunk)
            chunk_type = chunk['type']
            if chunk_type== "status":
                status_content = chunk['content']
                status_type = status_content['status_type']
                if status_type == "error":
                    error_msg = chunk.get('message')
                    self.logger.warning(f"XGAAreAgent execute_agent_loop: Fatal error - {error_msg}")
                    self.log_error(error_msg)
                elif status_type == "stop":
                    error_msg = chunk.get('message')
                    self.logger.warning("XGAAreAgent execute_agent_loop: Agent stopped.")
                    self.append_agent_log(
                        StopLog(
                            content     = f"Agent stopped - {error_msg}",
                            timestamp   = self.make_timestamp(),
                            agent_id    = self.agent_id,
                        )
                    )
            elif chunk_type == "assistant":
                llm_content  = chunk['content']['content']
                if "<thought>" in llm_content:
                    thought_content = re.search(r'<thought>(.*?)</thought>', llm_content, re.DOTALL).group(1).strip()
                    if thought_content:
                        self.append_agent_log(
                            ThoughtLog(
                                content     = thought_content,
                                timestamp   = self.make_timestamp(),
                                agent_id    = self.agent_id,
                            )
                        )
            elif chunk_type == "tool":
                tool_content = chunk['content']
                tool_execution = tool_content.get('tool_execution')
                tool_result = tool_execution.get('result')
                self.append_agent_log(
                    ObservationLog(
                        content     = str(tool_result),
                        attachments = [],
                        timestamp   = self.make_timestamp(),
                        agent_id    = self.agent_id,
                    )
                )

                self.append_agent_log(
                    ToolCallLog(
                        tool_name       = tool_execution.get('function_name'),
                        tool_arguments  = tool_execution.get('arguments'),
                        timestamp       = self.make_timestamp(),
                        agent_id        = self.agent_id,
                    )
                )
            #print(chunk)

        final_result = self.task_engine.parse_final_result(chunks)
        print(f"\n\nFINAL_RESULT: {final_result}")

        # Send Final Result to user
        args = {
            'content': final_result['content']
        }
        self.tools['AgentUserInterface__send_message_to_user'](**args)
        self.append_agent_log(
            ToolCallLog(
                tool_name       = 'AgentUserInterface__send_message_to_user',
                tool_arguments  = args,
                timestamp       = self.make_timestamp(),
                agent_id        = self.agent_id,
            )
        )

        # Return Final Result
        if final_result['type'] == "error":
            self.custom_state["running_state"] = RunningState.FAILED
        else:
            self.custom_state["running_state"] = RunningState.TERMINATED

        self.append_agent_log(
            FinalAnswerLog(
                content     = final_result['content'],
                timestamp   = self.make_timestamp(),
                agent_id    = self.agent_id
            )
        )



"""
XML Tool Call Parser Module

This module provides a reliable XML tool call parsing system that supports
the XML format with structured function_calls blocks.
"""

import json
import re
from dataclasses import dataclass
from typing import List, Dict, Any, Optional, Tuple

from xgae.utils import log_trace

@dataclass
class XMLToolCall:
    """Represents a parsed XML tool call."""
    function_name: str
    parameters: Dict[str, Any]
    raw_xml: str
    parsing_details: Dict[str, Any]


class XMLToolParser:
    """
    Parser for XML tool calls format:
    
    <function_calls>
    <invoke name="function_name">
    <parameter name="param_name">param_value</parameter>
    ...
    </invoke>
    </function_calls>
    """
    
    # Regex patterns for extracting XML blocks
    FUNCTION_CALLS_PATTERN = re.compile(
        r'<function_calls>(.*?)</function_calls>',
        re.DOTALL | re.IGNORECASE
    )
    
    INVOKE_PATTERN = re.compile(
        r'<invoke\s+name=["\']([^"\']+)["\']>(.*?)</invoke>',
        re.DOTALL | re.IGNORECASE
    )
    
    PARAMETER_PATTERN = re.compile(
        r'<parameter\s+name=["\']([^"\']+)["\']>(.*?)</parameter>',
        re.DOTALL | re.IGNORECASE
    )
    
    def __init__(self):
        """Initialize the XML tool parser."""
        pass
    
    def parse_content(self, content: str) -> List[XMLToolCall]:
        """
        Parse XML tool calls from content.
        
        Args:
            content: The text content potentially containing XML tool calls
            
        Returns:
            List of parsed XMLToolCall objects
        """
        tool_calls = []
        
        # Find function_calls blocks
        function_calls_matches = self.FUNCTION_CALLS_PATTERN.findall(content)
        
        for func_content in function_calls_matches:
            # Find all invoke blocks within this function_calls block
            invoke_matches = self.INVOKE_PATTERN.findall(func_content)
            
            for function_name, invoke_content in invoke_matches:
                try:
                    tool_call = self._parse_invoke_block(
                        function_name, 
                        invoke_content,
                        func_content
                    )
                    if tool_call:
                        tool_calls.append(tool_call)
                except Exception as e:
                    log_trace(e, f"XMLToolParser: Error parsing function={function_name}, invoke_content:\n{invoke_content}")
        
        return tool_calls
    
    def _parse_invoke_block(
        self, 
        function_name: str, 
        invoke_content: str,
        full_block: str
    ) -> Optional[XMLToolCall]:
        """Parse a single invoke block into an XMLToolCall."""
        parameters = {}
        parsing_details = {
            'function_name': function_name,
            'raw_parameters': {}
        }

        param_matches = self.PARAMETER_PATTERN.findall(invoke_content)
        for param_name, param_value in param_matches:
            param_value = param_value.strip()
            parsed_value = self._parse_parameter_value(param_value)
            parameters[param_name] = parsed_value
            parsing_details['raw_parameters'][param_name] = param_value
        
        # Extract the raw XML for this specific invoke
        invoke_pattern = re.compile(
            rf'<invoke\s+name=["\']{re.escape(function_name)}["\']>.*?</invoke>',
            re.DOTALL | re.IGNORECASE
        )
        raw_xml_match = invoke_pattern.search(full_block)
        raw_xml = raw_xml_match.group(0) if raw_xml_match else f"<invoke name=\"{function_name}\">...</invoke>"
        
        return XMLToolCall(
            function_name   = function_name,
            parameters      = parameters,
            raw_xml         = raw_xml,
            parsing_details = parsing_details
        )
    
    def _parse_parameter_value(self, value: str) -> Any:
        """
        Parse a parameter value, attempting to convert to appropriate type.
        
        Args:
            value: The string value to parse
            
        Returns:
            Parsed value (could be dict, list, bool, int, float, or str)
        """
        value = value.strip()
        
        # Try to parse as JSON first
        if value.startswith(('{', '[')):
            try:
                return json.loads(value)
            except json.JSONDecodeError:
                pass

        # Try to parse as string
        if value.startswith("\"") and value.endswith("\""):
            return value.strip("\"")

        #Try to parse as boolean
        if value.lower() in ('true', 'false'):
            return value.lower() == 'true'
        
        #Try to parse as number, this will be error when string type para
        try:
            if '.' in value:
                return float(value)
            else:
                return int(value)
        except ValueError:
            pass
        
        # Return as string
        return value


if __name__ == "__main__":
    xml_parser = XMLToolParser()
    content = "<function_calls>\\n<invoke name=\"RentAFlat__save_apartment\">\\n<parameter name=\"apartment_id\">\"+5674\"</parameter>\\n</invoke>\\n</function_calls>"
    xml_calls: List[XMLToolCall] = xml_parser.parse_content(content)
    print(str(xml_calls))


import json
import logging
import os

from typing import List, Any, Dict, Optional
from typing_extensions import override

from langchain_mcp_adapters.client import MultiServerMCPClient
from langchain_mcp_adapters.tools import load_mcp_tools

from xgae.engine.engine_base import XGAError, XGAToolSchema, XGAToolBox, XGAToolResult, XGAToolType
from xgae.tools.system_tools import XGASystemTools

class XGAMcpToolBox(XGAToolBox):
    GENERAL_MCP_SERVER_NAME = "XgaGeneral"
    A2A_MCP_SERVER_NAME = "XgaA2A"

    ST_SPLIT = "__"
    END_TASK_TOOL_NAME = "end_xgae_task"

    GENERAL_END_TASK_TOOL = GENERAL_MCP_SERVER_NAME + ST_SPLIT + END_TASK_TOOL_NAME
    A2A_END_TASK_TOOL = A2A_MCP_SERVER_NAME + ST_SPLIT + END_TASK_TOOL_NAME

    def __init__(self,
                 custom_mcp_server_file: Optional[str] = None,
                 custom_mcp_server_config: Optional[Dict[str, Any]] = None
        ):
        tool_box_mcp_server_config = []
        if custom_mcp_server_config:
            tool_box_mcp_server_config = custom_mcp_server_config
        elif custom_mcp_server_file:
            custom_mcp_server_config = self._load_mcp_servers_config(custom_mcp_server_file)
            tool_box_mcp_server_config = custom_mcp_server_config.get('mcpServers', {})

        if len(tool_box_mcp_server_config) > 0:
            self._mcp_client = MultiServerMCPClient(tool_box_mcp_server_config)

        self.mcp_server_names: List[str] = [server_name for server_name in tool_box_mcp_server_config]
        self.mcp_tool_schemas: Dict[str, List[XGAToolSchema]] = {}
        self.task_tool_schemas: Dict[str, Dict[str,XGAToolSchema]] = {}

        self._is_loaded_mcp_tool_schemas = False
        self.system_tools = XGASystemTools()

    @override
    async def init_tool_schemas(self):
        await self._load_mcp_tools_schema()


    @override
    async def creat_task_tool_box(self, task_id: str, general_tools: List[str], custom_tools: List[str]):
        system_tool_schemas = self.system_tools.get_sys_tool_schemas()
        task_tool_schemas = {tool_schema.tool_name: tool_schema for tool_schema in system_tool_schemas}

        general_tools = general_tools or []
        custom_tools = custom_tools or []

        general_tool_schemas = self.mcp_tool_schemas.get(self.GENERAL_MCP_SERVER_NAME, {})
        if "*" in general_tools:
            general_tool_schemas = {tool_schema.tool_name: tool_schema for tool_schema in general_tool_schemas}
            task_tool_schemas.update(general_tool_schemas)
        else:
            for tool_schema in general_tool_schemas:
                if tool_schema.function_name in general_tools:
                    task_tool_schemas[tool_schema.tool_name] = tool_schema

        if len(custom_tools) == 1 and custom_tools[0] == "*":
            custom_tools = []
            for server_name in self.mcp_server_names:
                if server_name != self.GENERAL_MCP_SERVER_NAME:
                    custom_tools.append(f"{server_name}.*")

        for server_tool_name in custom_tools:
            parts = server_tool_name.split(".")
            if len(parts) != 2:
                continue
            custom_server_name, custom_tool_name = parts
            if (not custom_server_name ) or (not custom_tool_name):
                continue

            custom_tool_schemas = self.mcp_tool_schemas.get(custom_server_name, None)
            if custom_tool_schemas is None:
                continue
            if custom_tool_name == "*":
                custom_tool_schema_dict = {tool_schema.tool_name: tool_schema for tool_schema in custom_tool_schemas}
                task_tool_schemas.update(custom_tool_schema_dict)
            else:
                for tool_schema in custom_tool_schemas:
                    if custom_tool_name == tool_schema.function_name:
                        task_tool_schemas[tool_schema.tool_name] = tool_schema

        task_tool_schemas.pop(self.GENERAL_END_TASK_TOOL, None)
        task_tool_schemas.pop(self.A2A_END_TASK_TOOL, None)

        self.task_tool_schemas[task_id] = task_tool_schemas


    @override
    async def destroy_task_tool_box(self, task_id: str):
        try:
            if len(self.get_task_tool_schemas(task_id, "general")) > 0:
                await self.call_tool(task_id, self.GENERAL_END_TASK_TOOL, {})
            if len(self.get_task_tool_schemas(task_id, "agent")) > 0:
                await self.call_tool(task_id, self.A2A_END_TASK_TOOL, {})
        except Exception as error:
            logging.error(f"McpToolBox destroy_task_tool_box: {error}")
        finally:
            self.task_tool_schemas.pop(task_id, None)


    @override
    def get_task_tool_names(self, task_id: str) -> List[str]:
        task_tool_schema = self.task_tool_schemas.get(task_id, {})
        task_tool_names =  list(task_tool_schema.keys())
        return task_tool_names


    @override
    def get_task_tool_schemas(self, task_id: str, tool_type: XGAToolType) -> List[XGAToolSchema]:
        task_tool_schemas = []

        all_task_tool_schemas = self.task_tool_schemas.get(task_id, {})
        for tool_schema in all_task_tool_schemas.values():
            if tool_schema.tool_type == tool_type:
                task_tool_schemas.append(tool_schema)

        return task_tool_schemas


    @override
    async def call_tool(self, task_id: str, tool_name: str, args: Optional[Dict[str, Any]] = None) -> XGAToolResult:
        if tool_name == self.GENERAL_END_TASK_TOOL:
            server_name = self.GENERAL_MCP_SERVER_NAME
            function_name = self.END_TASK_TOOL_NAME
        elif tool_name == self.A2A_END_TASK_TOOL:
            server_name = self.A2A_MCP_SERVER_NAME
            function_name = self.END_TASK_TOOL_NAME
        else:
            task_tool_schemas = self.task_tool_schemas.get(task_id, {})
            tool_schema = task_tool_schemas.get(tool_name, None)
            if tool_schema is None:
                raise XGAError(f"MCP tool not found: '{tool_name}'")
            server_name = tool_schema.server_name
            function_name = tool_schema.function_name

        tool_type = self._get_tool_type(server_name)
        if tool_type == "system":
            tool_args = args or {}
            tool_args = dict({'task_id': task_id}, **tool_args)
            result = self.system_tools.call_tool(tool_name, tool_args)
        else:
            async with self._mcp_client.session(server_name) as session:
                tools = await load_mcp_tools(session)
                mcp_tool = next((t for t in tools if t.name == function_name), None)

                if mcp_tool:
                    tool_args = args or {}
                    if tool_type == "general" or tool_type == "agent":
                        tool_args = dict({'task_id': task_id}, **tool_args)

                    try:
                        tool_result = await mcp_tool.arun(tool_args)
                        if tool_type == "general" or function_name == self.END_TASK_TOOL_NAME:
                            tool_result = json.loads(tool_result)
                            if  'success' in tool_result:
                                output = tool_result.get('output', "")
                                result = XGAToolResult(success=tool_result['success'], output=output)
                            else:
                                result = XGAToolResult(success=False, output=str(tool_result))
                        else:
                            result = XGAToolResult(success=True, output=str(tool_result))
                    except Exception as e:
                        error = f"Call mcp tool '{function_name}' error: {str(e)}"
                        logging.error(f"McpToolBox call_tool: {error}")
                        result = XGAToolResult(success=False, output=error)
                else:
                    error = f"No MCP tool found with name: {tool_name}"
                    logging.info(f"McpToolBox call_tool: error={error}")
                    result =  XGAToolResult(success=False, output=error)

        return result


    async def _load_mcp_tools_schema(self)-> None:
        if not self._is_loaded_mcp_tool_schemas:
            for server_name in self.mcp_server_names:
                self.mcp_tool_schemas[server_name] = []
                try:
                    mcp_tools = await self._mcp_client.get_tools(server_name=server_name)
                except Exception as e:
                    logging.error(f"### McpToolBox load_mcp_tools_schema: Langchain mcp get_tools failed, "
                                  f"need start mcp server '{server_name}' !")
                    continue

                tool_type = self._get_tool_type(server_name)
                for tool in mcp_tools:
                    input_schema = tool.args_schema
                    if tool_type == "general" or tool_type == "agent":
                        input_schema['properties'].pop("task_id", None)
                        if 'task_id' in input_schema['required']:
                            input_schema['required'].remove('task_id')
                        params_properties = input_schema.get('properties', {})
                        for param_properties in params_properties.values():
                            param_properties.pop('title', None)

                    metadata = tool.metadata or {}
                    tool_schema = XGAToolSchema(
                        tool_name       = server_name + self.ST_SPLIT + tool.name,
                        function_name   = tool.name,
                        tool_type       = tool_type,
                        server_name     = server_name,
                        description     = tool.description,
                        input_schema    = input_schema,
                        metadata        = metadata
                    )
                    self.mcp_tool_schemas[server_name].append(tool_schema)

            self._is_loaded_mcp_tool_schemas = True

    async def reload_mcp_tools_schema(self) -> None:
        self._is_loaded_mcp_tool_schemas = False
        await self.init_tool_schemas()


    def _load_mcp_servers_config(self, mcp_config_path: str) -> Dict[str, Any]:
        try:
            if os.path.exists(mcp_config_path):
                with open(mcp_config_path, 'r', encoding="utf-8") as f:
                    server_config = json.load(f)

                    for server_name, server_info in server_config['mcpServers'].items():
                        if "transport" not in server_info:
                            if "url" in server_info:
                                server_info['transport'] = "streamable_http" if "mcp" in server_info['url'] else "sse"
                            else:
                                server_info['transport'] = "stdio"

                    return server_config
            else:
                logging.warning(f"McpToolBox load_mcp_servers_config: MCP servers config file not found at: {mcp_config_path}")
                return {'mcpServers': {}}

        except Exception as e:
            logging.error(f"McpToolBox load_mcp_servers_config: Failed to load MCP servers config: {e}")
            return {'mcpServers': {}}


    def _get_tool_type(self, server_name: str) -> XGAToolType:
        tool_type: XGAToolType = "custom"

        if server_name == self.system_tools.SYSTEM_SERVER_NAME:
            tool_type = "system"
        elif server_name == self.GENERAL_MCP_SERVER_NAME:
            tool_type = "general"
        elif server_name == self.A2A_MCP_SERVER_NAME:
            tool_type = "agent"

        return tool_type

if __name__ == "__main__":
    import asyncio
    from dataclasses import asdict
    from xgae.utils.setup_env import setup_logging

    setup_logging()

    async def main():
        ## Before Run Exec: uv run example-fault-tools
        mcp_tool_box = XGAMcpToolBox(custom_mcp_server_file="xga_mcp_servers.json")
        #mcp_tool_box = XGAMcpToolBox()

        task_id = "task1"
        await mcp_tool_box.init_tool_schemas()
        #await mcp_tool_box.creat_task_tool_box(task_id=task_id, general_tools=["*"], custom_tools=["*"])
        await mcp_tool_box.creat_task_tool_box(task_id=task_id, general_tools=["web_search"], custom_tools=["BomcFault.get_alarm"])
        tool_types = ["system", "general", "custom",  "agent"]

        has_custom_tool = False
        has_agent_tool = False
        has_general_tool = False
        for tool_type in tool_types:
            tool_schemas = mcp_tool_box.get_task_tool_schemas(task_id, tool_type)
            if len(tool_schemas) > 0:
                print(f"{tool_type}_tools_schemas" + "*"*50)
            for tool_schema in tool_schemas:
                print(asdict(tool_schema))
            if len(tool_schemas) > 0 and tool_type == "custom":
                has_custom_tool = True
            if len(tool_schemas) > 0 and tool_type == "agent":
                has_agent_tool = True
            if len(tool_schemas) > 0 and tool_type == "general":
                has_general_tool = True
            print("")


        result = await mcp_tool_box.call_tool(task_id=task_id, tool_name="complete")
        print(f"SYSTEM TOOL: call complete result: {result}")

        if has_custom_tool:
            result = await mcp_tool_box.call_tool(task_id=task_id, tool_name="BomcFault__get_alarm", args={'ip' : "10.1.2.1"})
            print(f"CUSTOM TOOL: call get_alarm result: {result}")

        if has_general_tool:
            result = await mcp_tool_box.call_tool(task_id=task_id, tool_name="XgaGeneral__web_search", args={'query' : "today weather of beijing"})
            print(f"GENERAL TOOL: call web_search result: {result}")

        if has_agent_tool:
            result = await mcp_tool_box.call_tool(task_id=task_id, tool_name="XgaA2A__query_equip_fault_cause", args={'input' : "Fault code is F02"})
            print(f"A2A TOOL: call query_equip_fault_cause result: {result}")

        await mcp_tool_box.destroy_task_tool_box(task_id)

    asyncio.run(main())


import logging

from typing import List, Dict, Any, Optional, AsyncGenerator
from typing_extensions import override

from xgae.utils import log_trace

from xgae.engine.responser.responser_base import TaskResponseProcessor, TaskResponserContext, TaskRunContinuousState


class StreamTaskResponser(TaskResponseProcessor):
    def __init__(self, response_context: TaskResponserContext):
        super().__init__(response_context)

    @override
    async def process_response(self,
                               llm_response: AsyncGenerator,
                               prompt_messages: List[Dict[str, Any]],
                               continuous_state: TaskRunContinuousState
                               ) -> AsyncGenerator[Dict[str, Any], None]:
        accumulated_content = continuous_state['accumulated_content']
        auto_continue_count = continuous_state['auto_continue_count']
        can_auto_continue   = continuous_state['auto_continue']
        msg_sequence        = continuous_state['assistant_msg_sequence']

        use_assistant_chunk_msg = self.response_context["use_assistant_chunk_msg"]

        finish_reason = None
        should_auto_continue = False

        logging.info(f"=== StreamResp：Start Process Response, assistant_msg_sequence={msg_sequence}, "
                     f"accumulated_content_len={len(accumulated_content)}")
        try:
            async for llm_chunk in llm_response:
                if hasattr(llm_chunk, 'choices') and llm_chunk.choices and hasattr(llm_chunk.choices[0],'finish_reason'):
                    if llm_chunk.choices[0].finish_reason:
                        finish_reason = llm_chunk.choices[0].finish_reason # LLM finish reason: ‘stop' , 'length'
                        logging.info(f"StreamResp：LLM chunk response finish_reason={finish_reason}")

                if hasattr(llm_chunk, 'choices') and llm_chunk.choices:
                    llm_chunk_msg = llm_chunk.choices[0].delta if hasattr(llm_chunk.choices[0], 'delta') else None

                    if llm_chunk_msg and hasattr(llm_chunk_msg, 'content') and llm_chunk_msg.content:
                        chunk_content = llm_chunk_msg.content
                        accumulated_content += chunk_content

                        xml_tool_chunks = self._extract_xml_chunks(accumulated_content)
                        xml_tool_chunk_len = len(xml_tool_chunks)
                        if self.max_xml_tool_calls <= 0 or xml_tool_chunk_len <= self.max_xml_tool_calls:
                            if use_assistant_chunk_msg:
                                message_data = {"role": "assistant", "content": chunk_content}
                                metadata = {"sequence": msg_sequence}
                                assistant_chunk_msg = self.create_response_message(type="assistant_chunk",
                                                                                   content=message_data,
                                                                                   is_llm_message=True,
                                                                                   metadata=metadata)
                                yield assistant_chunk_msg
                                msg_sequence += 1
                        else:
                            finish_reason = "xml_tool_limit_reached"
                            logging.warning(f"StreamResp: Over XML Tool Limit, finish_reason='xml_tool_limit_reached', "
                                            f"xml_tool_chunk_len={xml_tool_chunk_len}")
                            break

            parsed_xml_data = self._parse_xml_tool_calls(accumulated_content)
            if finish_reason == "xml_tool_limit_reached":
                parsed_xml_data = parsed_xml_data[:self.max_xml_tool_calls]

            should_auto_continue = (can_auto_continue and finish_reason == 'length')

            self.root_span.event(name=f"stream_processor_start[{self.task_no}]({auto_continue_count})", level="DEFAULT",
                                 status_message=f"finish_reason={finish_reason}, tool_exec_strategy={self.tool_exec_strategy}, "
                                                f"parsed_xml_data_len={len(parsed_xml_data)}, accumulated_content_len={len(accumulated_content)}, "
                                                f"should_auto_continue={should_auto_continue}")

            assistant_msg_id = None
            if accumulated_content and not should_auto_continue:
                message_data = {"role": "assistant", "content": accumulated_content}
                assistant_msg = self.add_response_message(type="assistant", content=message_data, is_llm_message=True)
                yield assistant_msg
                assistant_msg_id = assistant_msg['message_id']

            tool_calls_to_execute = [item['tool_call'] for item in parsed_xml_data]
            if len(tool_calls_to_execute) > 0 and not should_auto_continue:
                tool_results = await self._execute_tools(tool_calls_to_execute, self.tool_exec_strategy)
                tool_index = 0
                for i, (returned_tool_call, tool_result) in enumerate(tool_results):
                    parsed_xml_item = parsed_xml_data[i]
                    tool_call = parsed_xml_item['tool_call']
                    parsing_details = parsed_xml_item['parsing_details']

                    tool_context = self._create_tool_context(tool_call, tool_index, assistant_msg_id, parsing_details, tool_result)

                    tool_start_msg = self._add_tool_start_message(tool_context)
                    yield tool_start_msg

                    tool_message = self._add_tool_messsage(tool_context, self.xml_adding_strategy)

                    tool_completed_msg = self._add_tool_completed_message(tool_context, tool_message['message_id'])
                    yield tool_completed_msg

                    yield tool_message

                    if tool_context.function_name in ['ask', 'complete']:
                        finish_reason = "completed"
                        break

                    tool_index += 1
            else:
                finish_reason = "non_tool_call"
                logging.warning(f"StreamResp: finish_reason='non_tool_call', No Tool need to call !")

            if finish_reason:
                finish_content = {'status_type': "finish", 'finish_reason': finish_reason}
                finish_msg = self.add_response_message(type="status", content=finish_content, is_llm_message=False)
                yield finish_msg
        except Exception as e:
            trace = log_trace(e, f"StreamResp: Process response accumulated_content:\n {accumulated_content}")
            self.root_span.event(name="stream_response_process_error", level="ERROR",
                                 status_message=f"Process streaming response error: {e}",
                                 metadata={"content": accumulated_content, "trace": trace})

            content = {'role': "system", 'status_type': "error", 'message': f"Process streaming response error: {e}"}
            error_msg = self.add_response_message(type="status", content=content, is_llm_message=False)
            yield error_msg

            raise  # Use bare 'raise' to preserve the original exception with its traceback
        finally:
            if should_auto_continue:
                continuous_state['accumulated_content'] = accumulated_content
                continuous_state['assistant_msg_sequence'] = msg_sequence
                logging.warning(f"StreamResp: Updated continuous state for auto-continue with {len(accumulated_content)} chars")


import logging


def log_trace(exception: Exception, error: str=None) -> str:
    import traceback

    if error:
        logging.error(f"{error} , error: {exception}")

    trace_info = traceback.format_exc()
    logging.error("Trace Details:\n%s", traceback.format_exc())

    return trace_info


def get_trace() -> str:
    import traceback

    return traceback.format_exc()


def to_bool(value: any) -> bool:
    if value is None:
        return False

    return str(value).lower() == "true"




import os
import litellm
import asyncio

from dotenv import load_dotenv
from litellm import acompletion, completion

load_dotenv()

env_public_key = os.getenv("LANGFUSE_PUBLIC_KEY")
env_secret_key = os.getenv("LANGFUSE_SECRET_KEY")
env_host = os.getenv("LANGFUSE_HOST", "https://cloud.langfuse.com")
print(f"Langfuse: env_public_key: {env_public_key}, env_secret_key: {env_secret_key}, env_host: {env_host}")

env_llm_model = os.getenv("LLM_MODEL", "openai/qwen3-235b-a22b")
env_llm_api_key = os.getenv("LLM_API_KEY")
env_llm_api_base = os.getenv("LLM_API_BASE", "https://dashscope.aliyuncs.com/compatible-mode/v1")
print(f"LLM: env_llm_model: {env_llm_model}, env_llm_api_key: {env_llm_api_key}, env_llm_api_base: {env_llm_api_base}")

if env_public_key and env_secret_key:
    litellm.success_callback = ["langfuse"]
    litellm.failure_callback = ["langfuse"]

    response = completion(
        model=env_llm_model,
        api_key=env_llm_api_key,
        api_base=env_llm_api_base,
        enable_thinking=False,
        stream=False,
        messages=[
            {"role": "user", "content": "2+3="}
        ],
        metadata={
            "generation_name": "litellm-completion-test",  # set langfuse generation name
        }
    )

    print("completion:" + response.choices[0].message.content)

async def llm_acompletion():
    response = await acompletion(
        model=env_llm_model,
        api_key=env_llm_api_key,
        api_base=env_llm_api_base,
        enable_thinking=False,
        stream=False,
        messages=[
            {"role": "user", "content": "4+5="}
        ],
        metadata={
            "generation_name": "litellm-acompletion-test",  # set langfuse generation name
        }
    )

    print("acompletion:"  + response.choices[0].message.content)


asyncio.run(llm_acompletion())


import logging
import os

from typing import Any, Dict, List, Optional, AsyncGenerator
from uuid import uuid4

from langfuse.callback import CallbackHandler
from langfuse import Langfuse

from langgraph.graph import END, START, StateGraph
from langgraph.types import interrupt, Command
from langgraph.checkpoint.memory import MemorySaver
from langgraph.config import get_stream_writer

from xgae.utils.misc import read_file
from xgae.utils import log_trace

from xgae.engine.engine_base import XGATaskResult
from xgae.engine.mcp_tool_box import XGAMcpToolBox
from xgae.engine.task_engine import XGATaskEngine

from examples.agent.langgraph.reflection.agent_base import AgentContext, TaskState, EvaluateResult
from examples.agent.langgraph.reflection.result_eval_agent import TaskResultEvalAgent
from examples.agent.langgraph.reflection.custom_prompt_rag import CustomPromptRag


class ReflectiontAgent:
    MAX_TASK_RETRY = 2
    QUALIFIED_RESULT_SCORE = 0.7

    def __init__(self, use_prompt_rag: Optional[bool] = False):
        self.graph = None

        self.graph_config = None
        self.graph_langfuse = None
        self.task_engine: XGATaskEngine = None

        self.tool_box = XGAMcpToolBox(custom_mcp_server_file="xga_mcp_servers.json")
        self.result_eval_agent = TaskResultEvalAgent()

        if use_prompt_rag:
            self.custom_prompt_rag = CustomPromptRag()

    async def _create_graph(self) -> StateGraph:
        try:
            graph_builder = StateGraph(TaskState)

            # Add nodes
            graph_builder.add_node('supervisor', self._supervisor_node)
            graph_builder.add_node('prompt_optimize', self._prompt_optimize_node)
            graph_builder.add_node('select_tool', self._select_tool_node)
            graph_builder.add_node('exec_task', self._exec_task_node)
            graph_builder.add_node('final_result', self._final_result_node)

            # Add edges
            graph_builder.add_edge(START, 'supervisor')
            graph_builder.add_conditional_edges(
                'supervisor',
                self._next_condition,
                {
                    'select_tool'       : 'select_tool',
                    'exec_task'         : 'exec_task',
                    'prompt_optimize'   : 'prompt_optimize',
                    'end'               : END
                }
            )

            graph_builder.add_edge('prompt_optimize', 'select_tool')
            graph_builder.add_edge('select_tool', 'exec_task')
            graph_builder.add_edge('exec_task', 'final_result')

            graph_builder.add_conditional_edges(
                'final_result',
                self._next_condition,
                {
                    'supervisor': 'supervisor',
                    'exec_task' : 'exec_task',
                    'end'       : END
                }
            )
            
            graph = graph_builder.compile(checkpointer=MemorySaver())
            graph.name = "XGARectAgentGraph"

            return graph
        except Exception as e:
            logging.error("Failed to create XGARectAgent Graph: %s", str(e))
            raise


    def _search_system_prompt(self, user_input: str) -> str:
        system_prompt = None
        if hasattr(self, 'custom_prompt_rag'):
            system_prompt_path = self.custom_prompt_rag.search_prompt(user_input)
            if system_prompt_path:
                system_prompt = read_file(system_prompt_path)
        else:
            if "fault" in user_input:  # only for example
                system_prompt = read_file("templates/example/fault_user_prompt.md")
        return system_prompt


    async def _supervisor_node(self, state: TaskState) -> Dict[str, Any]:
        user_input = state['user_inputs'][0]
        eval_result = state.get('eval_result', None)
        system_prompt = state.get('system_prompt', None)

        if system_prompt is  None and eval_result is None:
            system_prompt = self._search_system_prompt(user_input)
        is_system_prompt = True if system_prompt is not None else False

        general_tools = [] if system_prompt else ["*"]
        custom_tools = ["*"] if system_prompt  else []

        task_plan_score = None
        if eval_result and 'task_plan' in eval_result and 'score' in eval_result['task_plan']:
            task_plan_score = eval_result['task_plan'].get('score', 1.0)

        function_call_score = None
        if eval_result and 'function_call' in eval_result and 'score' in eval_result['function_call']:
            function_call_score = eval_result['function_call'].get('score', 1.0)

        super_state = {}
        if task_plan_score and task_plan_score < self.QUALIFIED_RESULT_SCORE:
            next_node = "prompt_optimize"
            super_state = self._prepare_task_retry(state)
            logging.warning(f"****** ReactAgent TASK_RETRY: task_plan_score={task_plan_score} < {self.QUALIFIED_RESULT_SCORE} , "
                            f"Start Optimize Prompt ...")
        elif function_call_score and function_call_score < self.QUALIFIED_RESULT_SCORE:
            next_node = "select_tool"
            super_state = self._prepare_task_retry(state)
            logging.warning(f"****** ReactAgent TASK_RETRY: function_call_score={function_call_score} < {self.QUALIFIED_RESULT_SCORE} , "
                            f"Select Tool Again ...")
        elif eval_result is not None:  # retry condition is not satisfied, end task
            next_node = "end"
        else:
            next_node = "select_tool" if is_system_prompt else "exec_task"

        logging.info(f"ReactAgent supervisor_node: is_system_prompt={is_system_prompt}, next_node={next_node}")

        super_state['next_node']        = next_node
        super_state['system_prompt']    = system_prompt
        super_state['custom_tools']     = custom_tools
        super_state['general_tools']    = general_tools

        return super_state


    async def _prompt_optimize_node(self, state: TaskState) -> Dict[str, Any]:
        system_prompt = state['system_prompt']
        logging.info("ReactAgent prompt_optimize_node: optimize system prompt")
        # @todo optimize system prompt in future
        return {
            'system_prompt' : system_prompt,
        }


    def _select_custom_tools(self, system_prompt: str) -> list[str]:
        # @todo search mcp tool based on system prompt or user_input in future
        custom_tools = ["*"] if system_prompt  else []
        return custom_tools


    async def _select_tool_node(self, state: TaskState) -> Dict[str, Any]:
        system_prompt = state.get('system_prompt',None)
        general_tools = []

        logging.info("ReactAgent select_tool_node: select tool based on system_prompt")
        custom_tools = self._select_custom_tools(system_prompt)
        return {
            'general_tools' : general_tools,
            'custom_tools'  : custom_tools,
        }


    async def _exec_task_node(self, state: TaskState) -> Dict[str, Any]:
        user_input = state['user_inputs'][0]
        system_prompt = state.get('system_prompt',None)
        general_tools = state.get('general_tools',[])
        custom_tools = state.get('custom_tools',[])
        retry_count = state.get('retry_count', 0)
        task_no = state.get('task_no', 0)
        is_system_prompt = True if system_prompt is not None else False

        trace_id = self.graph_langfuse.get_trace_id()
        llm_messages = []
        try:
            logging.info(f"🔥 ReactAgent exec_task_node: user_input={user_input}, general_tools={general_tools}, "
                         f"custom_tools={custom_tools}, is_system_prompt={is_system_prompt}")

            # if langgraph resume , must use same task engine
            if self.task_engine is None:
                self.task_engine = XGATaskEngine(
                    task_id         = state['agent_context']['task_id'],
                    task_no         = task_no,
                    session_id      = state['agent_context'].get('session_id', None),
                    user_id         = state['agent_context'].get('user_id', None),
                    agent_id        = state['agent_context'].get('agent_id', None),
                    tool_box        = self.tool_box,
                    general_tools   = general_tools,
                    custom_tools    = custom_tools,
                    system_prompt   = system_prompt
                )
                retry_count += 1

            chunks = []
            stream_writer = get_stream_writer()
            async for chunk in self.task_engine.run_task(task_input={"role": "user", "content": user_input},
                                                         trace_id=trace_id):
                chunks.append(chunk)
                stream_writer({"engine_message": chunk})

            task_result = self.task_engine.parse_final_result(chunks)
            llm_messages = self.task_engine.get_history_llm_messages()
            task_no += 1  # a task use unique task_no, no matter retry n times
        except Exception as e:
            logging.error(f"XReactAgent exec_task_node: Failed to execute task: {e}")
            task_result = XGATaskResult(type="error", content="Failed to execute task")

        return {
            'task_result'   : task_result,
            'retry_count'   : retry_count,
            'llm_messages'  : llm_messages.copy(),
            'task_no'       : task_no,
        }


    async def _final_result_node(self, state: TaskState) -> Dict[str, Any]:
        user_inputs = state['user_inputs']
        task_result = state['task_result']
        llm_messages = state['llm_messages']
        agent_context = state['agent_context']
        system_prompt = state.get('system_prompt', None)
        retry_count = state['retry_count']

        is_system_prompt = True if system_prompt is not None else False

        next_node = "end"
        final_result = task_result
        eval_result = None
        if task_result['type'] == "ask":
            logging.info(f"XReactAgent final_result_node: ASK_USER_QUESTION: {task_result['content']}")
            ask_input = interrupt({
                'final_result' : task_result
            })
            logging.info(f"XReactAgent final_result_node: ASK_USER_ANSWER: {ask_input}")
            next_node = "exec_task"
            user_inputs.insert(0, ask_input)
            final_result = None
        elif is_system_prompt and retry_count < self.MAX_TASK_RETRY:
            trace_id = self.graph_langfuse.get_trace_id()
            session_id = agent_context.get('session_id', None)
            task_input = ", ".join(reversed(user_inputs))
            eval_result = await self.result_eval_agent.eval_result(task_input, system_prompt, task_result,
                                                                   llm_messages, trace_id, session_id)
            if 'task_result' in eval_result and 'score' in eval_result['task_result']:
                score =  eval_result['task_result'].get('score', 1.0)
                if score < self.QUALIFIED_RESULT_SCORE:
                    next_node = "supervisor"
        
        logging.info(f"ReactAgent final_result_node: next_node={next_node}")
        
        return {
            'user_inputs'    : user_inputs,
            'next_node'     : next_node,
            'final_result'  : final_result,
            'eval_result'   : eval_result
        }


    def _next_condition(self, state: TaskState) -> str:
        next_node = state['next_node']
        return next_node


    async def generate_with_result(self, user_input: str,
                                   agent_context: Optional[AgentContext] = None,
                                   is_resume: Optional[bool]=False) -> XGATaskResult:
        agent_context = agent_context or {}
        try:

            if is_resume:
                logging.info(f"=== Start React Agent for USER_ASK_ANSWER: {user_input}")
                graph_input = Command(resume=user_input)
            else:
                logging.info(f"=== Start React Agent for USER_INPUT: {user_input}")
                graph_input = await self._prepare_graph_start(user_input, agent_context)

            final_state = await self.graph.ainvoke(graph_input, config=self.graph_config)

            if "__interrupt__" in final_state:
                interrupt_event = final_state["__interrupt__"][0]
                interrupt_value = interrupt_event.value
                result = interrupt_value['final_result']
            else:
                result = final_state['final_result']

            return result
        except Exception as e:
            log_trace(e, f"XReactAgent generate: user_input={user_input}")
            result = XGATaskResult(type="error", content=f"React Agent error: {e}")
            return result


    async def generate(self, user_input: str,
                       agent_context: Optional[AgentContext]=None,
                       is_resume: Optional[bool]=False) -> AsyncGenerator[Dict[str, Any], None]:
        agent_context = agent_context or {}
        try:
            if is_resume:
                logging.info(f"=== Start React Stream Agent for USER_ASK_ANSWER: {user_input}")
                graph_input = Command(resume=user_input)
            else:
                logging.info(f"=== Start React Stream Agent USER_ASK_ANSWER: {user_input}")
                graph_input = await self._prepare_graph_start(user_input, agent_context)

            async for msg_type, message in self.graph.astream(input=graph_input,
                                                              config=self.graph_config,
                                                              stream_mode=["custom", "updates"]):
                if msg_type == "updates" and '__interrupt__' in message:
                    interrupt_event = message["__interrupt__"][0]
                    interrupt_value = interrupt_event.value
                    final_result = interrupt_value['final_result']
                    yield final_result
                elif msg_type == "updates" and 'final_result' in message:
                    message = message['final_result']
                    final_result = message.get('final_result', None)
                    if final_result:
                        yield final_result
                elif msg_type == "custom" and 'engine_message' in message:
                    message = {'type': "message", 'content': message['engine_message']}
                    yield message

        except Exception as e:
            log_trace(e, f"XReactAgent generate: user_input={user_input}")
            yield {'type': "error", 'content': f"React Agent generate error: {e}"}


    async def _prepare_graph_start(self, user_input, agent_context: AgentContext)->TaskState:
        if self.graph is None:
            self.graph = await self._create_graph()

        self._clear_graph()

        agent_context = agent_context or {}
        task_id = agent_context.get("task_id", f"xga_task_{uuid4()}")
        agent_context["task_id"] = task_id
        thread_id = agent_context.get('thread_id', task_id)
        agent_context['thread_id'] = thread_id
        session_id = agent_context.get('session_id', task_id)
        agent_context['session_id'] = session_id


        langfuse_handler = self._get_langfuse_handler(agent_context)
        callbacks = None
        if langfuse_handler:
            callbacks = [langfuse_handler]
            self.graph_langfuse = langfuse_handler.langfuse
        else:
            self.graph_langfuse = Langfuse(enabled=False)

        self.graph_config = {
            'recursion_limit': 100,
            'configurable': {
                'thread_id': thread_id
            },
            'callbacks': callbacks
        }

        graph_input = {
            'user_inputs'       : [user_input],
            'next_node'         : None,
            'agent_context'     : agent_context,
            'retry_count'       : 0,
            'task_no'           : 0
        }

        return graph_input


    def _get_langfuse_handler(self, agent_context: AgentContext)->CallbackHandler:
        langfuse_handler = None
        public_key = os.getenv("LANGFUSE_PUBLIC_KEY")
        secret_key = os.getenv("LANGFUSE_SECRET_KEY")
        host = os.getenv("LANGFUSE_HOST", "https://cloud.langfuse.com")

        if public_key and secret_key:
            langfuse_handler =  CallbackHandler(
                public_key  = public_key,
                secret_key  = secret_key,
                host        = host,
                trace_name  = "xga_react_agent",
                session_id  = agent_context.get('session_id', None),
                user_id     = agent_context.get('user_id', None),
            )
        return langfuse_handler


    def _clear_graph(self):
        self.graph_config = None
        self.graph_langfuse = None
        self.task_engine: XGATaskEngine = None


    def _prepare_task_retry(self, state: TaskState)-> Dict[str, Any]:
        self.task_engine = None
        user_inputs = state['user_inputs']
        task_input = ", ".join(reversed(user_inputs))

        return {
            'user_inputs'   : [task_input],
            'llm_messages'  : [],
            'task_result'   : None,
            'final_result'  : None,
            'eval_result'   : None,
        }


from typing import Union, Optional, Dict, List, Any, Literal, TypedDict
from dataclasses import dataclass
from abc import ABC, abstractmethod

class XGAError(Exception):
    """Custom exception for errors in the XGA system."""
    pass

XGAMsgStatusType = Literal["error", "stop", "finish", "tool_started", "tool_completed", "tool_error", "tool_failed"]
XGAResponseMsgType = Literal["user", "status",  "tool", "assistant", "assistant_chunk"]

class XGAResponseMessage(TypedDict, total=False):
    message_id: str
    type: XGAResponseMsgType
    is_llm_message: bool
    content: Union[Dict[str, Any], str]
    metadata: Dict[str, Any]

class XGATaskResult(TypedDict, total=False):
    type: Literal["ask", "answer", "error"]
    content: str
    attachments: Optional[List[str]]

XGAToolType = Literal["system", "general", "custom",  "agent"]

@dataclass
class XGAToolSchema:
    tool_name: str
    function_name: str
    tool_type: XGAToolType
    server_name: str
    description: str
    input_schema: Dict[str, Any]
    metadata: Optional[Dict[str, Any]]


@dataclass
class XGAToolResult:
    success: bool
    output: str


class XGAToolBox(ABC):
    @abstractmethod
    async def init_tool_schemas(self):
        pass

    @abstractmethod
    async def creat_task_tool_box(self, task_id: str, general_tools: List[str], custom_tools: List[str]):
        pass

    @abstractmethod
    async def destroy_task_tool_box(self, task_id: str):
        pass

    @abstractmethod
    def get_task_tool_schemas(self, task_id: str, type: XGAToolType) -> List[XGAToolSchema]:
        pass

    @abstractmethod
    def get_task_tool_names(self, task_id: str) -> List[str]:
        pass

    @abstractmethod
    async def call_tool(self, task_id: str, tool_name: str, args: Optional[Dict[str, Any]] = None) -> XGAToolResult:
        pass


import json
import logging
import re

from typing import Any, Dict, List, Optional

from xgae.engine.engine_base import XGATaskResult
from xgae.utils.misc import read_file
from xgae.utils.llm_client import LLMClient, LangfuseMetadata


class TaskResultEvalAgent:
    def __init__(self):
        self.model_client = LLMClient()
        self.prompt_template: str = read_file("templates/example/result_eval_template.txt")


    async def eval_result(self,
                          task_input: str,
                          task_plan: str,
                          task_result: XGATaskResult,
                          llm_messages: List[Dict[str, Any]],
                          trace_id: Optional[str] = None,
                          session_id: Optional[str] = None)-> Dict[str, Any]:
        prompt = self._build_prompt(task_input, task_plan, task_result, llm_messages)
        messages = [{"role": "user", "content": prompt}]

        langfuse_metadata = self._create_llm_langfuse_meta(trace_id, session_id)

        logging.info("TaskResultEvalAgent: LLM Eval result.....")
        response = await self.model_client.acompletion(messages, langfuse_metadata)
        response_text = await self.model_client.get_acompletion_response(response)

        cleaned_text = re.sub(r'^\s*```json|```\s*$', '', response_text, flags=re.MULTILINE).strip()
        eval_result = json.loads(cleaned_text)

        result_score = eval_result.get('task_result', {}).get('score', -1)
        plan_score = eval_result.get('task_plan', {}).get('score', -1)
        function_score = eval_result.get('function_call', {}).get('score', -1)

        logging.info(f"FINAL_RESULT_SCORE: task_result_score={result_score}, "
                     f"task_plan_score={plan_score}, function_call_score={function_score}")
        return eval_result


    def _build_prompt(self, task_input: str, task_plan: str, task_result: XGATaskResult, llm_messages: List[Dict[str, Any]])-> str:
        prompt = self.prompt_template.replace("{task_input}", task_input)
        prompt = prompt.replace("{task_result}", str(task_result))
        llm_process = ""
        function_process = ""
        llm_step = 1
        function_step = 1
        for llm_message in llm_messages:
            content = llm_message.get('content', '')
            if "tool_execution" in content:
                function_process += f"{function_step}. \n"
                tool_exec = json.loads(content)
                func_call = tool_exec['tool_execution']
                func_call.pop('xml_tag_name')
                clear_content = json.dumps(func_call, indent=2)
                function_process += clear_content
                function_process += "\n"
                function_step += 1
            else:
                llm_process += f"{llm_step}. \n"
                llm_process += content
                llm_process += "\n"
                llm_step += 1

        prompt = prompt.replace("{task_plan}", task_plan)
        prompt = prompt.replace("{llm_process}", llm_process)
        prompt = prompt.replace("{function_process}", function_process)

        return prompt


    def _create_llm_langfuse_meta(self, trace_id:str, session_id: str)-> LangfuseMetadata:
        generation_name = "xga_agent_final_result_completion"

        return LangfuseMetadata(
            generation_name     = generation_name,
            existing_trace_id   = trace_id,
            session_id          = session_id
        )



if __name__ == "__main__":
    import asyncio
    from xgae.utils.setup_env import setup_logging
    setup_logging()

    async def main():
        final_result_agent = TaskResultEvalAgent()

        task_plan = read_file("templates/example/fault_user_prompt.md")
        user_input = "locate 10.2.3.4 fault and solution"

        answer = ("Task Summary: The fault for IP 10.2.3.4 was identified as a Business Recharge Fault (Code: F01), "
                  "caused by a Phone Recharge Application Crash. The solution applied was to restart the application. "
                  "Key Deliverables: Fault diagnosis and resolution steps. Impact Achieved: Service restored.")
        task_result:XGATaskResult = {'type': "answer",  'content': answer}

        llm_messages: List[Dict[str, Any]] = [{
            'content':
                    """<function_calls> 
                        <invoke name="get_alarm_type">
                                       <parameter name="alarm_id">alm0123</parameter>
                         </invoke>
                    </function_calls>'""",
             'role': "assistant"
             },{
            'content': """{"tool_execution": {
                                "function_name": "get_alarm_type", 
                                "xml_tag_name": "get-alarm-type", 
                                "arguments": {"alarm_id": "alm0123"}, 
                                "result": {"success": true, "output": "1", "error": null}}}""",
            'role': 'assistant'
            }]

        return await final_result_agent.eval_result(user_input, task_plan, task_result, llm_messages)


    final_result = asyncio.run(main())
    final_result_json = json.dumps(final_result, ensure_ascii=False, indent=2)
    print(f"FINAL_RESULT：   {final_result_json} ")


import asyncio
import json
import logging

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Dict, Any, Optional, Tuple, Union, Literal, Callable, TypedDict, AsyncGenerator

from xgae.utils import log_trace
from xgae.utils.json_helpers import safe_json_parse
from xgae.utils.xml_tool_parser import XMLToolParser

from xgae.engine.engine_base import XGAToolResult, XGAToolBox
from xgae.engine.task_langfuse import XGATaskLangFuse


# Type alias for XML result adding strategy
XmlAddingStrategy = Literal["user_message", "assistant_message", "inline_edit"]

# Type alias for tool execution strategy
ToolExecutionStrategy = Literal["sequential", "parallel"]

class TaskResponserContext(TypedDict, total=False):
    is_stream: bool
    task_id: str
    task_run_id: str
    task_no: int
    model_name: str
    max_xml_tool_calls: int             # LLM generate max_xml_tool limit, 0 is no limit
    use_assistant_chunk_msg: bool
    tool_exec_strategy: ToolExecutionStrategy
    xml_adding_strategy: XmlAddingStrategy
    add_response_msg_func: Callable
    create_response_msg_func: Callable
    tool_box: XGAToolBox
    task_langfuse: XGATaskLangFuse


class TaskRunContinuousState(TypedDict, total=False):
    accumulated_content: str
    auto_continue_count: int
    auto_continue: bool
    assistant_msg_sequence: int


@dataclass
class ToolExecutionContext:
    """Context for a tool execution including call details, result, and display info."""
    tool_call: Dict[str, Any]
    tool_index: int
    function_name: str
    xml_tag_name: str
    result: Optional[XGAToolResult] = None
    error: Optional[Exception] = None
    assistant_message_id: Optional[str] = None
    parsing_details: Optional[Dict[str, Any]] = None


class TaskResponseProcessor(ABC):
    def __init__(self, response_context: TaskResponserContext):
        self.response_context = response_context

        self.task_id                = response_context['task_id']
        self.task_run_id            = response_context['task_run_id']
        self.task_no                = response_context['task_no']
        self.tool_exec_strategy     = response_context['tool_exec_strategy']
        self.xml_adding_strategy    = response_context['xml_adding_strategy']
        self.max_xml_tool_calls     = response_context['max_xml_tool_calls']

        self.add_response_message       = response_context['add_response_msg_func']
        self.create_response_message    = response_context['create_response_msg_func']
        self.tool_box                   = response_context['tool_box']

        task_langfuse = response_context['task_langfuse']
        self.root_span = task_langfuse.root_span

        self.xml_parser                 = XMLToolParser()



    @abstractmethod
    async def process_response(self,
                               llm_response: AsyncGenerator,
                               prompt_messages: List[Dict[str, Any]],
                               continuous_state: TaskRunContinuousState
                               ) -> AsyncGenerator[Dict[str, Any], None]:
        pass


    def _extract_xml_chunks(self, content: str) -> List[str]:
        """Extract complete XML chunks using start and end pattern matching."""
        chunks = []
        pos = 0

        try:
            # First, look for new format <function_calls> blocks
            start_pattern = '<function_calls>'
            end_pattern = '</function_calls>'

            while pos < len(content):
                # Find the next function_calls block
                start_pos = content.find(start_pattern, pos)
                if start_pos == -1:
                    break

                # Find the matching end tag
                end_pos = content.find(end_pattern, start_pos)
                if end_pos == -1:
                    break

                # Extract the complete block including tags
                chunk_end = end_pos + len(end_pattern)
                chunk = content[start_pos:chunk_end]
                chunks.append(chunk)

                # Move position past this chunk
                pos = chunk_end

            # If no new format found, fall back to old format for backwards compatibility
            if not chunks:
                pos = 0
                while pos < len(content):
                    # Find the next tool tag
                    next_tag_start = -1
                    current_tag = None

                    # Find the earliest occurrence of any registered tool function name
                    # Check for available function names
                    available_func_names = self.tool_box.get_task_tool_names(self.task_id)
                    for func_name in available_func_names:
                        # Convert function name to potential tag name (underscore to dash)
                        tag_name = func_name.replace('_', '-')
                        start_pattern = f'<{tag_name}'
                        tag_pos = content.find(start_pattern, pos)

                        if tag_pos != -1 and (next_tag_start == -1 or tag_pos < next_tag_start):
                            next_tag_start = tag_pos
                            current_tag = tag_name

                    if next_tag_start == -1 or not current_tag:
                        break

                    # Find the matching end tag
                    end_pattern = f'</{current_tag}>'
                    tag_stack = []
                    chunk_start = next_tag_start
                    current_pos = next_tag_start

                    while current_pos < len(content):
                        # Look for next start or end tag of the same type
                        next_start = content.find(f'<{current_tag}', current_pos + 1)
                        next_end = content.find(end_pattern, current_pos)

                        if next_end == -1:  # No closing tag found
                            break

                        if next_start != -1 and next_start < next_end:
                            # Found nested start tag
                            tag_stack.append(next_start)
                            current_pos = next_start + 1
                        else:
                            # Found end tag
                            if not tag_stack:  # This is our matching end tag
                                chunk_end = next_end + len(end_pattern)
                                chunk = content[chunk_start:chunk_end]
                                chunks.append(chunk)
                                pos = chunk_end
                                break
                            else:
                                # Pop nested tag
                                tag_stack.pop()
                                current_pos = next_end + 1

                    if current_pos >= len(content):  # Reached end without finding closing tag
                        break

                    pos = max(pos + 1, current_pos)
        except Exception as e:
            trace = log_trace(e, f"TaskProcessor extract_xml_chunks: Error extracting XML chunks: {content}")
            self.root_span.event(name="task_process_extract_xml_chunk_error", level="ERROR",
                                 status_message=f"Error extracting XML chunks: {e}",
                                 metadata={"content": content, "trace": trace})

        return chunks

    def _parse_xml_tool_call(self, xml_chunk: str) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Parse XML chunk into tool call format and return parsing details.

        Returns:
            Tuple of (tool_call, parsing_details) or None if parsing fails.
            - tool_call: Dict with 'function_name', 'xml_tag_name', 'arguments'
            - parsing_details: Dict with 'attributes', 'elements', 'text_content', 'root_content'
        """
        try:
            # Check if this is the new format (contains <function_calls>)
            if '<function_calls>' in xml_chunk and '<invoke' in xml_chunk:
                # Use the new XML parser
                parsed_calls = self.xml_parser.parse_content(xml_chunk)

                if not parsed_calls:
                    logging.error(f"TaskProcessor parse_xml_tool_call: No tool calls found in XML chunk: {xml_chunk}")
                    return None

                # Take the first tool call (should only be one per chunk)
                xml_tool_call = parsed_calls[0]
                if not xml_tool_call.function_name:
                    logging.error(f"TaskProcessor parse_xml_tool_call: xml_tool_call function name is empty: {xml_tool_call}")
                    return None

                # Convert to the expected format
                tool_call = {
                    'function_name' : xml_tool_call.function_name,
                    'xml_tag_name'  : xml_tool_call.function_name.replace("_", "-"),  # For backwards compatibility
                    'arguments'     : xml_tool_call.parameters
                }

                # Include the parsing details
                parsing_details = xml_tool_call.parsing_details
                parsing_details['raw_xml'] = xml_tool_call.raw_xml

                logging.debug(f"TaskProcessor parse_xml_tool_call: Parsed new format tool call: {tool_call}")
                return tool_call, parsing_details

            # If not the expected <function_calls><invoke> format, return None
            logging.error(f"TaskProcessor parse_xml_tool_call: XML chunk does not contain expected <function_calls><invoke> format: {xml_chunk}")
        except Exception as e:
            trace = log_trace(e, f"TaskProcessor parse_xml_tool_call: Error parsing XML chunk: {xml_chunk}")
            self.root_span.event(name="task_process_parsing_xml_chunk_error", level="ERROR",
                                 status_message=f"Error parsing XML chunk: {e}",
                                 metadata={"xml_chunk": xml_chunk, "trace": trace})
            return None

    def _parse_xml_tool_calls(self, content: str) -> List[Dict[str, Any]]:
        """Parse XML tool calls from content string.

        Returns:
            List of dictionaries, each containing {'tool_call': ..., 'parsing_details': ...}
        """
        parsed_data = []
        xml_chunk = None
        try:
            xml_chunks = self._extract_xml_chunks(content)

            for xml_chunk in xml_chunks:
                result = self._parse_xml_tool_call(xml_chunk)
                if result:
                    tool_call, parsing_details = result
                    parsed_data.append({
                        "tool_call": tool_call,
                        "parsing_details": parsing_details
                    })
        except Exception as e:
            trace = log_trace(e, f"TaskProcessor parse_xml_tool_calls: Error parsing XML tool calls, xml_chunk: {xml_chunk}")
            self.root_span.event(name="task_process_parse_xml_tool_calls_error", level="ERROR",
                             status_message=f"Error parsing XML tool calls: {e}",
                                 metadata={"content": xml_chunk, "trace": trace})

        return parsed_data


    async def _execute_tool(self, tool_call: Dict[str, Any]) -> XGAToolResult:
        """Execute a single tool call and return the result."""
        function_name = tool_call['function_name']
        exec_tool_span = self.root_span.span(name=f"execute_tool.{function_name}", input=tool_call["arguments"])
        try:
            arguments = tool_call.get('arguments', {})
            if isinstance(arguments, str):
                try:
                    arguments = safe_json_parse(arguments)
                except json.JSONDecodeError:
                    logging.warning(f"TaskProcessor execute_tool: Tool '{function_name}' arguments is not dict type, args={arguments}")
                    arguments = {"text": arguments}  # useless

            result = None
            available_tool_names = self.tool_box.get_task_tool_names(self.task_id)
            if function_name in available_tool_names:
                logging.info(f"TaskProcessor execute_tool: Tool '{function_name}' executing, args={arguments}")
                result = await self.tool_box.call_tool(self.task_id, function_name, arguments)
            else:
                logging.error(f"TaskProcessor execute_tool: Tool function '{function_name}' not found in toolbox")
                result = XGAToolResult(success=False, output=f"Tool function '{function_name}' not found")

            logging.info(f"TaskProcessor execute_tool: Tool '{function_name}' execution complete, result: {result}")
            exec_tool_span.update(status_message="tool_executed", output=result)

            return result
        except Exception as e:
            trace = log_trace(e, f"TaskProcessor execute_tool: Executing tool {function_name}")

            exec_tool_span.update(status_message="task_process_tool_exec_error", level="ERROR",
                                  output=f"Error executing tool {function_name}, error: {str(e)}",
                                  metadata={"trace": trace})

            return XGAToolResult(success=False, output=f"Executing tool {function_name}, error: {str(e)}")


    async def _execute_tools(self, tool_calls: List[Dict[str, Any]],
                             execution_strategy: ToolExecutionStrategy = "sequential"
                             ) -> List[Tuple[Dict[str, Any], XGAToolResult]]:
        if execution_strategy == "sequential":
            return await self._execute_tools_sequentially(tool_calls)
        elif execution_strategy == "parallel":
            return await self._execute_tools_in_parallel(tool_calls)
        else:
            logging.warning(f"TaskProcessor execute_tools: Unknown execution strategy '{execution_strategy}', use sequential")
            return await self._execute_tools_sequentially(tool_calls)


    async def _execute_tools_sequentially(self, tool_calls: List[Dict[str, Any]]) -> List[Tuple[Dict[str, Any], XGAToolResult]]:
        """Execute tool calls sequentially and return results.

        This method executes tool calls one after another, waiting for each tool to complete
        before starting the next one. This is useful when tools have dependencies on each other.

        Returns:
            List of tuples containing the original tool call and its result
        """
        if not tool_calls:
            logging.warning("TaskProcessor execute_tools_sequentially: tool_calls is empty")
            return []

        tool_names = [tc['function_name'] for tc in tool_calls]
        tool_num = len(tool_calls)
        if tool_num > 1:
            logging.info(f"TaskProcessor execute_tools_sequentially: Executing {tool_num} tools sequentially: {tool_names}")
            self.root_span.event(name="task_process_execute_tools_sequentially", level="DEFAULT",
                                  status_message=f"Executing {len(tool_calls)} tools sequentially: {tool_names}")

        results = []
        for index, tool_call in enumerate(tool_calls):
            tool_name = tool_call['function_name']
            logging.info(f"TaskProcessor execute_tools_sequentially: Executing tool '{tool_name}', sequence={index + 1}/{tool_num}")
            result = await self._execute_tool(tool_call)
            results.append((tool_call, result))

            # Check if this is a terminating tool (ask or complete)
            if tool_name in ["ask", "complete"]:
                if len(results) < tool_num:
                    logging.info(f"TaskProcessor execute_tools_sequentially: Terminating tool '{tool_name}' executed, Stopping further tool execution.")
                    self.root_span.event(name="task_process_terminate_tool_executed", level="DEFAULT",
                                         status_message=f"Terminating tool '{tool_name}' executed, Stopping further tool execution.")
                break

        logging.info(f"TaskProcessor execute_tools_sequentially: Execution completed for {len(results)} tools, total {tool_num} tools)")
        return results


    async def _execute_tools_in_parallel(self, tool_calls: List[Dict[str, Any]]) -> List[Tuple[Dict[str, Any], XGAToolResult]]:
        """Execute tool calls in parallel and return results.

        This method executes all tool calls simultaneously using asyncio.gather, which
        can significantly improve performance when executing multiple independent tools.

        Returns:
            List of tuples containing the original tool call and its result
        """
        if not tool_calls:
            logging.warning("TaskProcessor execute_tools_in_parallel: tool_calls is empty")
            return []

        tool_names = [tc['function_name'] for tc in tool_calls]
        tool_num = len(tool_calls)
        if tool_num > 1:
            logging.info(f"TaskProcessor execute_tools_in_parallel: Executing {tool_num} tools sequentially: {tool_names}")
            self.root_span.event(name="task_process_execute_tools_parallel", level="DEFAULT",
                                  status_message=f"Executing {len(tool_calls)} tools parallelly: {tool_names}")

        # Execute all tasks concurrently with error handling
        tasks = [self._execute_tool(tool_call) for tool_call in tool_calls]
        results = await asyncio.gather(*tasks, return_exceptions=True)

        processed_results = []
        for i, (tool_call, result) in enumerate(zip(tool_calls, results)):
            processed_results.append((tool_call, result))

        logging.info(f"TaskProcessor execute_tools_in_parallel: Execution completed for {len(results)} tools, total {tool_num} tools)")
        return processed_results

    def _create_tool_context(self,
                             tool_call: Dict[str, Any],
                             tool_index: int,
                             assistant_message_id: Optional[str] = None,
                             parsing_details: Optional[Dict[str, Any]] = None,
                             result: Optional[XGAToolResult] = None,
                             ) -> ToolExecutionContext:
        """Create a tool execution context with display name and parsing details populated."""
        return ToolExecutionContext(
            tool_call               = tool_call,
            tool_index              = tool_index,
            function_name           = tool_call['function_name'],
            xml_tag_name            = tool_call['xml_tag_name'],
            assistant_message_id    = assistant_message_id,
            parsing_details         = parsing_details,
            result                  = result
        )


    def _add_tool_messsage(self,context:ToolExecutionContext, strategy: XmlAddingStrategy) -> Optional[Dict[str, Any]]:  # Return the full message object
        # Create two versions of the structured result
        # Rich version for the frontend
        result_for_frontend = self._create_structured_tool_result(context.tool_call, context.result, context.parsing_details, for_llm=False)
        # Concise version for the LLM
        result_for_llm = self._create_structured_tool_result(context.tool_call, context.result, context.parsing_details, for_llm=True)

        # Add the message with the appropriate role to the conversation history
        # This allows the LLM to see the tool result in subsequent interactions
        role = "user" if strategy == "user_message" else "assistant"
        content = {
            'role': role,
            'content': json.dumps(result_for_llm)
        }

        metadata = {}
        if context.assistant_message_id:
            metadata['assistant_message_id'] = context.assistant_message_id

        if context.parsing_details:
            metadata['parsing_details'] = context.parsing_details

        metadata['frontend_content'] = result_for_frontend

        tool_message =  self.add_response_message(type="tool", content=content, is_llm_message=True, metadata=metadata)

        # Let's result_for_frontend the message for yielding.
        yield_message = tool_message.copy()
        yield_message['content'] = result_for_frontend

        return yield_message


    def _create_structured_tool_result(self,
                                       tool_call: Dict[str, Any],
                                       result: XGAToolResult,
                                       parsing_details: Optional[Dict[str, Any]] = None,
                                       for_llm: bool = False) -> Dict[str, Any]:
        function_name = tool_call['function_name']
        xml_tag_name = tool_call['xml_tag_name']
        arguments = tool_call.get('arguments', {})

        # Process the output - if it's a JSON string, parse it back to an object
        output = result.output
        parsed_output = safe_json_parse(output)
        if isinstance(parsed_output, (dict, list)):
            output = parsed_output

        output_to_use = output
        # If this is for the LLM and it's an edit_file tool, create a concise output
        if for_llm and function_name == 'edit_file' and isinstance(output, dict):
            # The frontend needs original_content and updated_content to render diffs.
            # The concise version for the LLM was causing issues.
            # @todo We will now pass the full output, and rely on the ContextManager to truncate if needed.
            output_to_use = output

        structured_result = {
            'tool_execution': {
                'function_name' : function_name,
                'xml_tag_name'  : xml_tag_name,
                'arguments'     : arguments,
                'result'      : {
                    'success' : result.success,
                    'output'  : output_to_use,
                    'error'   : None if result.success else result.output
                },
            }
        }

        return structured_result


    def _add_tool_start_message(self, context: ToolExecutionContext) -> Optional[Dict[str, Any]]:
        """Formats, saves, and returns a tool started status message."""
        content = {
            'status_type'       : "tool_started",
            'role'              : "assistant",
            'function_name'     : context.function_name,
            'xml_tag_name'      : context.xml_tag_name,
            'message'           : f"Starting execution of {context.function_name}",
            'tool_index'        : context.tool_index
        }

        return  self.add_response_message(type="status", content=content, is_llm_message=False)

    def _add_tool_completed_message(self, context: ToolExecutionContext, tool_message_id: Optional[str]) -> Optional[Dict[str, Any]]:
        """Formats, saves, and returns a tool completed/failed status message."""
        if not context.result:
            return  self._add_tool_error_message(context)

        status_type = "tool_completed" if context.result.success else "tool_failed"
        message_text = f"Tool {context.function_name} {'completed successfully' if context.result.success else 'failed'}"

        content = {
            'status_type'       : status_type,
            'role'              : "assistant",
            'function_name'     : context.function_name,
            'xml_tag_name'      : context.xml_tag_name,
            'message'           : message_text,
            'tool_index'        : context.tool_index
        }

        metadata = {}
        if tool_message_id:
            metadata['tool_result_message_id'] = tool_message_id

        return  self.add_response_message(type="status", content=content, is_llm_message=False, metadata=metadata)

    def _add_tool_error_message(self, context: ToolExecutionContext) -> Optional[Dict[str, Any]]:
        """Formats, saves, and returns a tool error status message."""
        error_msg = str(context.error) if context.error else "Tool execution unknown exception"
        content = {
            'status_type'       : "tool_error",
            'role'              : "assistant",
            'function_name'     : context.function_name,
            'xml_tag_name'      : context.xml_tag_name,
            'message'           : f"Executing tool {context.function_name} exception: {error_msg}",
            'tool_index'        : context.tool_index
        }

        return  self.add_response_message(type="status", content=content, is_llm_message=False)




import logging
import json
import os

from typing import List, Any, Dict, Optional, AsyncGenerator, Union, Literal
from uuid import uuid4

from xgae.utils import log_trace, to_bool
from xgae.utils.llm_client import LLMClient, LLMConfig


from xgae.engine.engine_base import XGAResponseMsgType, XGAResponseMessage, XGAToolBox, XGATaskResult
from xgae.engine.task_langfuse import XGATaskLangFuse
from xgae.engine.prompt_builder import XGAPromptBuilder
from xgae.engine.mcp_tool_box import XGAMcpToolBox
from xgae.engine.responser.responser_base import TaskResponserContext, TaskResponseProcessor, TaskRunContinuousState

class XGATaskEngine:
    def __init__(self,
                 task_id: Optional[str] = None,
                 task_no: Optional[int] = None,
                 session_id: Optional[str] = None,
                 user_id: Optional[str] = None,
                 agent_id: Optional[str] = None,
                 general_tools: Optional[List[str]] = None,
                 custom_tools: Optional[List[str]] = None,
                 system_prompt: Optional[str] = None,
                 max_auto_run: Optional[int] = None,
                 tool_exec_parallel: Optional[bool] = None,
                 llm_config: Optional[LLMConfig] = None,
                 prompt_builder: Optional[XGAPromptBuilder] = None,
                 tool_box: Optional[XGAToolBox] = None
                 ):
        self.task_id    = task_id if task_id else f"xga_task_{uuid4()}"
        self.session_id = session_id
        self.user_id    = user_id
        self.agent_id   = agent_id

        self.llm_client = LLMClient(llm_config)
        self.model_name = self.llm_client.model_name
        self.is_stream  = self.llm_client.is_stream

        self.prompt_builder                 = prompt_builder or XGAPromptBuilder(system_prompt)
        self.tool_box: XGAToolBox           = tool_box or XGAMcpToolBox()

        self.general_tools:List[str] = general_tools
        self.custom_tools:List[str]  = custom_tools

        max_auto_run = max_auto_run if max_auto_run  else int(os.getenv('MAX_AUTO_RUN', 15))
        self.max_auto_run: int = 1 if max_auto_run <= 1 else max_auto_run

        self.use_assistant_chunk_msg = to_bool(os.getenv('USE_ASSISTANT_CHUNK_MSG', False))
        self.tool_exec_parallel = True if tool_exec_parallel is None else tool_exec_parallel

        self.task_no = (task_no - 1) if task_no  else -1
        self.task_run_id :str = None
        self.task_prompt :str = None
        self.task_langfuse: XGATaskLangFuse = None

        self.task_response_msgs: List[XGAResponseMessage] = []

        self.terminate_task = False

    async def run_task_with_final_answer(self,
                                         task_input: Dict[str, Any],
                                         trace_id: Optional[str] = None) -> XGATaskResult:
        final_result: XGATaskResult = None
        try:
            await self._init_task()

            self.task_langfuse.start_root_span("run_task_with_final_answer", task_input, trace_id)

            chunks = []
            async for chunk in self.run_task(task_input, trace_id):
                chunks.append(chunk)

            if len(chunks) > 0:
                final_result = self.parse_final_result(chunks)
            else:
                final_result = XGATaskResult(type="error", content="LLM Answer is Empty")

            return final_result
        finally:
            self.task_langfuse.end_root_span("run_task_with_final_answer", final_result)


    async def run_task(self,
                       task_input: Dict[str, Any],
                       trace_id: Optional[str] = None) -> AsyncGenerator[Dict[str, Any], None]:
        try:
            await self._init_task()

            self.task_langfuse.start_root_span("run_task", task_input, trace_id)

            self.add_response_message(type="user", content=task_input, is_llm_message=True)

            async for chunk in self._run_task_auto():
                yield chunk
        finally:
            await self.tool_box.destroy_task_tool_box(self.task_id)
            self.task_langfuse.end_root_span("run_task")
            self.task_run_id = None

    async def _init_task(self) -> None:
        if self.task_run_id is None:
            self.task_no = self.task_no + 1
            self.task_run_id = f"{self.task_id}[{self.task_no}]"

            self.task_langfuse =self._create_task_langfuse()

            await  self.tool_box.init_tool_schemas()
            await self.tool_box.creat_task_tool_box(self.task_id, self.general_tools, self.custom_tools)

            system_tool_schemas = self.tool_box.get_task_tool_schemas(self.task_id, "system")
            general_tool_schemas = self.tool_box.get_task_tool_schemas(self.task_id, "general")
            custom_tool_schemas = self.tool_box.get_task_tool_schemas(self.task_id, "custom")
            agent_tool_schemas = self.tool_box.get_task_tool_schemas(self.task_id, "agent")

            self.task_prompt = self.prompt_builder.build_task_prompt(self.model_name,
                                                                     system_tool_schemas,
                                                                     general_tool_schemas,
                                                                     custom_tool_schemas,
                                                                     agent_tool_schemas)

            logging.info("*" * 10 + f"   XGATaskEngine Task'{self.task_id}' Initialized   " + "*" * 10)
            logging.info(f"model_name={self.model_name}, is_stream={self.is_stream}")
            logging.info(f"general_tools={self.general_tools}, custom_tools={self.custom_tools}")


    async def _run_task_auto(self) -> AsyncGenerator[Dict[str, Any], None]:
        continuous_state: TaskRunContinuousState = {
            'accumulated_content'   : "",
            'auto_continue_count'   : 0,
            'auto_continue'         : False if self.max_auto_run <= 1 else True,
            'assistant_msg_sequence': 0
        }

        def update_continuous_state(_auto_continue_count,  _auto_continue):
            continuous_state['auto_continue_count'] = _auto_continue_count
            continuous_state['auto_continue'] = _auto_continue

        auto_continue_count = 0
        no_tool_call_count = 0
        auto_continue = True
        while auto_continue and auto_continue_count < self.max_auto_run:
            auto_continue = False
            iterations = auto_continue_count

            try:
                async for chunk in self._run_task_once(continuous_state):
                    yield chunk
                    try:
                        if chunk['type'] == "status":
                            status_content = chunk['content']
                            status_type = status_content['status_type']
                            if status_type == "error":
                                logging.error(f"XGATaskEngine run_task_auto: task_response error: {chunk.get('message')}")
                                auto_continue = False
                                break
                            elif status_type == "finish":
                                finish_reason = status_content['finish_reason']
                                if finish_reason == "completed":
                                    logging.info(f"XGATaskEngine run_task_auto: Detected finish_reason='completed', TASK_COMPLETE Success !")
                                    auto_continue = False
                                    break
                                elif finish_reason == "xml_tool_limit_reached":
                                    logging.warning(f"XGATaskEngine run_task_auto: Detected finish_reason='xml_tool_limit_reached', stop auto-continue")
                                    auto_continue = False
                                    break
                                elif finish_reason == "non_tool_call":
                                    logging.warning(f"XGATaskEngine run_task_auto: Detected finish_reason='non_tool_call', stop auto-continue")
                                    if no_tool_call_count < 2: # Allow no tool call, for LLM summary or think, then call complete or other tool
                                        no_tool_call_count  += 1
                                        auto_continue = True
                                    else:
                                        auto_continue = False
                                        break
                                elif finish_reason in ["stop", "length"]: # 'length'  occur on some LLM
                                    auto_continue_count += 1
                                    no_tool_call_count = 0
                                    auto_continue = True if auto_continue_count < self.max_auto_run else False
                                    update_continuous_state(auto_continue_count, auto_continue)
                                    logging.info(f"XGATaskEngine run_task_auto: Detected finish_reason='{finish_reason}', auto-continuing ({auto_continue_count}/{self.max_auto_run})")
                    except Exception as parse_error:
                        trace = log_trace(parse_error,f"XGATaskEngine run_task_auto: Parse chunk error, chunk: {chunk}")
                        self.task_langfuse.root_span.event(name="engine_parse_chunk_error", level="ERROR",
                                             status_message=f"Task Engine parse chunk error: {parse_error}",
                                             metadata={"content": chunk, "trace": trace})

                        status_content = {'status_type': "error", 'role': "system", 'message': "Parse response chunk Error"}
                        error_msg = self.add_response_message(type="status", content=status_content, is_llm_message=False)
                        yield error_msg
            except Exception as run_error:
                trace = log_trace(run_error, "XGATaskEngine run_task_auto: Call task_run_once")
                self.task_langfuse.root_span.event(name="engine_task_run_once_error", level="ERROR",
                                                   status_message=f"Call task_run_once error: {run_error}",
                                                   metadata={"trace": trace})
                status_content = {'status_type': "error", 'role': "system", 'message': "Call run_task_once error"}
                error_msg = self.add_response_message(type="status", content=status_content, is_llm_message=False)
                yield error_msg
            finally:
                if not self.running_task_checkpoint("termination_check", iterations):
                    status_content = {'status_type': "stop", 'role': "system", 'message': "Task is termiated by Stop Command"}
                    error_msg = self.add_response_message(type="status", content=status_content, is_llm_message=False)
                    yield error_msg
                    break

    async def _run_task_once(self, continuous_state: TaskRunContinuousState) -> AsyncGenerator[Dict[str, Any], None]:
        llm_messages = [{"role": "system", "content": self.task_prompt}]
        cxt_llm_contents = self.get_history_llm_messages()
        llm_messages.extend(cxt_llm_contents)

        partial_content = continuous_state.get('accumulated_content', '')
        if partial_content:
            temp_assistant_message = {
                "role": "assistant",
                "content": partial_content
            }
            llm_messages.append(temp_assistant_message)

        auto_count = continuous_state.get("auto_continue_count")

        self.running_task_checkpoint("before_completion", auto_count, llm_messages)

        langfuse_metadata = self.task_langfuse.create_llm_langfuse_meta(auto_count)

        llm_response = await self.llm_client.acompletion(llm_messages, langfuse_metadata)
        response_processor = self._create_response_processer()

        async for chunk in response_processor.process_response(llm_response, llm_messages, continuous_state):
            self._logging_reponse_chunk(chunk, auto_count)

            if chunk['type'] == "assistant":
                assis_content = chunk['content']
                self.running_task_checkpoint("after_completion", auto_count, llm_messages, assis_content)

            yield chunk

    def parse_final_result(self, chunks: List[Dict[str, Any]]) -> XGATaskResult:
        final_result: XGATaskResult = None
        reverse_chunks = reversed(chunks)
        chunk = None

        # if self.terminate_task:
        #     return XGATaskResult(type="error", content="LLM Task is terminated !")

        try:
            finish_reason = ''
            for chunk in reverse_chunks:
                chunk_type = chunk['type']
                if chunk_type == "status":
                    status_content = chunk['content']
                    status_type = status_content['status_type']
                    if status_type == "error" or  status_type == "stop":
                        error = status_content['message']
                        final_result = XGATaskResult(type="error", content=error)
                    elif status_type == "finish":
                        finish_reason = status_content['finish_reason']
                elif chunk_type == "tool" and finish_reason in ['completed', 'stop', 'xml_tool_limit_reached', 'length']:
                    tool_content= chunk['content']
                    tool_execution = tool_content.get('tool_execution')
                    tool_name = tool_execution.get('function_name')
                    if tool_name == "complete":
                        result_content = tool_execution['arguments'].get('text', "Task completed with no answer")
                        attachments = tool_execution['arguments'].get('attachments', None)
                        final_result = XGATaskResult(type="answer", content=result_content, attachments=attachments)
                    elif tool_name == "ask":
                        result_content = tool_execution["arguments"].get("text", "Task ask for more info")
                        attachments = tool_execution["arguments"].get("attachments", None)
                        final_result = XGATaskResult(type="ask", content=result_content, attachments=attachments)
                    else:
                        # finish reason 1) 'stop': auto run reach max_auto_run limit 2) 'xml_tool_limit_reached' 3) 'length': occur on some LLM
                        tool_result = tool_execution.get("result", None)
                        if tool_result is not None:
                            success = tool_result['success']
                            output = tool_result.get('output', '')
                            result_type = "answer" if success else "error"
                            result_content = output
                            final_result = XGATaskResult(type=result_type, content=result_content)
                elif chunk_type == "assistant" and finish_reason == "non_tool_call":
                    assis_content = chunk['content']
                    result_content = assis_content.get('content', "LLM output is empty")
                    final_result = XGATaskResult(type="answer", content=result_content)

                if final_result:
                    break

            if final_result and (finish_reason == "completed" or finish_reason == "non_tool_call"):
                logging.info(f"✅ FINAL_RESULT: finish_reason={finish_reason}, final_result={final_result}")
            elif final_result is not None:
                logging.warning(f"⚠️ FINAL_RESULT: finish_reason={finish_reason}, final_result={final_result}")
            else:
                logging.warning(f"❌ FINAL_RESULT: LLM Result is EMPTY, finish_reason={finish_reason}")
                final_result = XGATaskResult(type="error", content="LLM has no answer")
        except Exception as e:
            trace = log_trace(e, f"XGATaskEngine parse_final_result: Parse message chunk error, chunk: {chunk}")
            self.task_langfuse.root_span.event(name="engine_parse_final_result_error", level="ERROR",
                                               status_message=f"Task Engine parse final result error: {e}",
                                               metadata={"content": chunk, "trace": trace})

            final_result = XGATaskResult(type="error", content="Parse final result failed!")

        return final_result


    def stop_task(self):
        logging.warning(f"⚠️ Begin Terminate Task: {self.task_id}")
        self.task_langfuse.root_span.event(name="stop_task", level="DEFAULT",
                                           status_message="Begin Terminate Task")
        self.terminate_task = True


    def running_task_checkpoint(self,
                                task_state: Literal["before_completion", "after_completion", "termination_check"],
                                iterations: int,
                                llm_messages: List[Dict[str, Any]] = None,
                                llm_response: Dict[str, Any] = None
                                )-> bool:
        if self.terminate_task and task_state == "termination_check":
            logging.warning(f"⚠️ TASK: {self.task_id} STOP RUNNING for STOP Command !")
        return not self.terminate_task


    def create_response_message(self, type: XGAResponseMsgType,
                             content: Union[Dict[str, Any], List[Any], str],
                             is_llm_message: bool,
                             metadata: Optional[Dict[str, Any]]=None)-> XGAResponseMessage:
        metadata = metadata or {}
        metadata['task_id']     = self.task_id
        metadata['task_run_id'] = self.task_run_id
        metadata['trace_id']    = self.task_langfuse.trace_id
        metadata['session_id']  = self.session_id
        metadata['user_id']     = self.user_id
        metadata['agent_id']    = self.agent_id

        message = XGAResponseMessage(
            message_id      = f"xga_msg_{uuid4()}",
            type            = type,
            is_llm_message  = is_llm_message,
            content         = content,
            metadata        = metadata
        )

        return message

    def add_response_message(self, type: XGAResponseMsgType,
                             content: Union[Dict[str, Any], List[Any], str],
                             is_llm_message: bool,
                             metadata: Optional[Dict[str, Any]]=None)-> XGAResponseMessage:
        message = self.create_response_message(type, content, is_llm_message, metadata)
        self.task_response_msgs.append(message)
        return message

    def get_history_llm_messages (self) -> List[Dict[str, Any]]:
        llm_messages = []
        for message in self.task_response_msgs:
            if message['is_llm_message'] and message['type'] != "assistant_chunk":
                llm_messages.append(message)

        response_llm_contents = []
        for llm_message in llm_messages:
            content = llm_message['content']
            if isinstance(content, str):
                try:
                    _content = json.loads(content)
                    response_llm_contents.append(_content)
                except json.JSONDecodeError as e:
                    pass
            else:
                response_llm_contents.append(content)

        return response_llm_contents

    def _create_response_processer(self) -> TaskResponseProcessor:
        response_context = self._create_response_context()
        is_stream = response_context['is_stream']
        if is_stream:
            from xgae.engine.responser.stream_responser import StreamTaskResponser
            return StreamTaskResponser(response_context)
        else:
            from xgae.engine.responser.non_stream_responser import NonStreamTaskResponser
            return NonStreamTaskResponser(response_context)

    def _create_response_context(self) -> TaskResponserContext:
        response_context: TaskResponserContext = {
            'is_stream'                 : self.is_stream,
            'task_id'                   : self.task_id,
            'task_run_id'               : self.task_run_id,
            'task_no'                   : self.task_no,
            'model_name'                : self.model_name,
            'max_xml_tool_calls'        : 0,
            'use_assistant_chunk_msg'   : self.use_assistant_chunk_msg,
            'tool_exec_strategy'        : "parallel" if self.tool_exec_parallel else "sequential",
            'xml_adding_strategy'       : "assistant_message", # user_message
            'add_response_msg_func'     : self.add_response_message,
            'create_response_msg_func'  : self.create_response_message,
            'tool_box'                  : self.tool_box,
            'task_langfuse'             : self.task_langfuse,
        }
        return response_context


    def _create_task_langfuse(self)-> XGATaskLangFuse:
        return XGATaskLangFuse(
            task_id      = self.task_id,
            task_run_id  = self.task_run_id,
            task_no      = self.task_no,
            session_id   = self.session_id,
            agent_id     = self.agent_id,
            user_id      = self.user_id
        )


    def _logging_reponse_chunk(self, chunk, auto_count: int)-> None:
        try:
            chunk_type = chunk['type']
            prefix = ""
            if chunk_type == "status":
                status_content = chunk['content']
                status_type = status_content.get('status_type', "empty")
                if status_type in ["tool_started", "tool_completed"]:
                    return
                prefix = "-" + status_type
            elif chunk_type == "tool":
                tool_content = chunk['content']
                tool_execution = tool_content.get('tool_execution')
                tool_name = tool_execution.get('function_name')
                prefix = "-" + tool_name

            status_content = chunk['content']
            pretty_content = status_content
            if isinstance(status_content, dict):
                pretty_content = json.dumps(status_content, ensure_ascii=False, indent=2)

            if chunk_type == "assistant_chunk":
                logging.debug(f"TASK_RESP_CHUNK[{self.task_no}]({auto_count})<{chunk_type}{prefix}> content: {pretty_content}")
            else:
                logging.info(f"TASK_RESP_CHUNK[{self.task_no}]({auto_count})<{chunk_type}{prefix}> content: {pretty_content}")
        except Exception as e:
            logging.error(f"XGATaskEngine logging_reponse_chunk: Decorate chunk={chunk}, error: {e}")



if __name__ == "__main__":
    import asyncio
    from xgae.utils.misc import read_file
    from xgae.utils.setup_env import setup_logging

    setup_logging()

    async def main():
        # Before Run Exec: uv run example-fault-tools
        tool_box = XGAMcpToolBox(custom_mcp_server_file="xga_mcp_servers.json")
        system_prompt = read_file("templates/example/fault_user_prompt.md")
        # system_prompt = None  # no business prompt, always can run success with good LLM
        engine =  XGATaskEngine(tool_box=tool_box,
                                custom_tools=["*"],
                                system_prompt=system_prompt,
                                session_id="session_1",
                                agent_id="agent_1"
                                )
        user_input = "locate 10.0.0.1 fault and solution"
        final_result = await engine.run_task_with_final_answer(task_input={'role': "user", 'content': user_input})
        print(f"FINAL RESULT:{final_result}")


    asyncio.run(main())


import chromadb

chroma_client = chromadb.EphemeralClient()

# default use 'ONNXMiniLM_L6_V2' embedding function
collection = chroma_client.get_or_create_collection(name="fault_collection")

collection.upsert(
    documents=[
        "Fault location and analysis of fault causes",
        "This is a analyse about alarm"
    ],
    ids=["id1", "id2"],
    metadatas=[{"type": 1}, {"type": 2}]
)

querys = ["locate fault and solution",
          "alarm search"
          ]

for query in querys:
    results = collection.query(
        query_texts=query,
        n_results=2,
        #where={"type": 1},
        where={"type": {"$gt": 0}},
        where_document = {"$contains": "fault"}
    )
    print("*"*50)
    print(f"query='{query}', id={results['ids'][0]}, score={results['distances'][0]}")




import logging
import os
from typing import  List
from typing_extensions import override

from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStore

from langchain_community.vectorstores import Chroma
from chromadb.utils.embedding_functions import OpenAIEmbeddingFunction

class ChromaEmbedding(Embeddings):
    embedding_model_name = "text-embedding-v3"

    def __init__(self):
        api_key = os.getenv('LLM_API_KEY')
        api_base = os.getenv('LLM_API_BASE', "https://dashscope.aliyuncs.com/compatible-mode/v1")

        self.embedding_function = OpenAIEmbeddingFunction(
            api_key = api_key,
            api_base = api_base,
            model_name = self.embedding_model_name,
        )


    @override
    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        return self.embedding_function(texts)


    @override
    def embed_query(self, text: str) -> List[float]:
        return self.embed_documents([text])[0]


class CustomPromptRag:
    PROMPT_SIMILAR_SCORE = 0.85   # need tune-up score based on different embedding model

    def __init__(self):
        prompt_docs = self._load_prompts()
        self.vector_store = self._init_vector_store(prompt_docs)


    # should read from DB, load all custom prompt or COT
    def _load_prompts(self) -> List[Document]:
        prompt_docs = []
        prompt_docs.append(self._create_prompt_doc(
            prompt_summary="Fault location and analysis of fault causes",
            prompt_path="templates/example/fault_user_prompt.md"
        ))
        return prompt_docs


    def _create_prompt_doc(self, prompt_summary: str, prompt_path: str)-> Document:
        return Document(
            page_content=prompt_summary,
            metadata={
                "source": prompt_path,
            }
        )


    def _init_vector_store(self, docs: List[Document]) -> VectorStore:
        embeddings = ChromaEmbedding()
        return Chroma.from_documents(
            documents=docs,
            embedding=embeddings,
            persist_directory=None
        )


    def search_prompt(self, query:str)-> str:
        prompt_path = None
        results = self.vector_store.similarity_search_with_score(query=query, k=1)
        if results and len(results) > 0:
            doc, score = results[0]
            if score > self.PROMPT_SIMILAR_SCORE:
                logging.info(f"CustomPromptRag search: SIMILAR_SCORE: {score} > {self.PROMPT_SIMILAR_SCORE}, "
                             f"\nquery: '{query}' \nprompt_summary: '{doc.page_content}'\n")
            else:
                prompt_path = doc.metadata['source']
                logging.info(f"CustomPromptRag search: SIMILAR_SCORE: {score}, prompt_path: '{prompt_path}'")

        return prompt_path


if __name__ == "__main__":
    from xgae.utils.setup_env import setup_logging

    setup_logging()

    custom_prompt_rag = CustomPromptRag()

    querys = ["locate 10.2.3.4 fault and solution", # 0.79
              "定位 10.2.3.4 故障,并给出解决方案",      # 0.81
              "locate fault and solution",          # 0.42
              "locate fault",                       # 0.40
              "定位故障",                            # 0.64
              "fault solution",                     # 0.47
              "locate",                             # 0.95
              "5+7"                                 # 1.12
              ]

    for query in querys:
        logging.info("*"*50)
        logging.info(f"query: '{query}'")
        custom_prompt_rag.search_prompt(query)



import asyncio

from xgae.engine.task_engine import XGATaskEngine

from xgae.utils.setup_env import setup_logging

setup_logging()

async def main() -> None:
    engine =  XGATaskEngine()

    final_result = await engine.run_task_with_final_answer(
        task_input={"role": "user", "content": "1+7"}
    )

    print("FINAL RESULT:", final_result)


asyncio.run(main())